/*
 * Copyright (c) 2008 Q2S NTNU, Trondheim, Norway
 *  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Port of the Swarm module developed by O. Wittner for ns-2.30.
 * Author: Laurent Paquereau <laurent.paquereau@q2s.ntnu.no>
 *
 * $Header$
 */

#ifndef _ceas_adib_h
#define _ceas_adib_h

#include <map>
#include <values.h>
#include "ceas-temperature.h"
#ifndef stand_alone
#include <config.h>
#include <stl-utils.h>
#include "ceas-logger.h"
#include "ceas-network-layer-unit.h"
#include "ceas-routing-unit-ib.h"

// Forward declarations
template<class _Key,class _PathItemData> class CEASRoutingUnit;
template<class _Key,class _CEASTemperature,class _PathItemData> class CEAS_ADIBEntry;
#else
template<class _CEASTemperature> class CEAS_ADIBEntry;
#endif


#ifndef stand_alone
//! At Destination Information Base
template<class __Key,class __PathItemData>
class CEAS_ADIB
: public CEASRoutingUnitIB<__Key,CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>*,__PathItemData> {
  public:
    CEAS_ADIB(CEASRoutingUnit<__Key,__PathItemData>* routingUnit);
    virtual CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>* addEntry(const __Key& key);
    inline CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>* operator[] (const __Key& key)
      {return getEntry(key);}
    inline unsigned int D() const {return D_;}
    inline double rho() const {return rho_;}
    inline double rhoReductionFactor() const {return rhoReductionFactor_;}
  protected:
    using CEASRoutingUnitIB<__Key,CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>*,__PathItemData>::
      base_;
    using CEASRoutingUnitIB<__Key,CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>*,__PathItemData>::
      routingUnit_;
    // Initial values of the parameters (set using CEAS-config)
    unsigned int D_; //!< control parameter used for rho reduction (#iterations)
    double rho_; //!< search focus
    double rhoReductionFactor_; //!< rho reduction Factor
    CEASTemperature_t temperatureType_; //!< type of temperature
};

//! Constructor
template<class __Key,class __PathItemData>
CEAS_ADIB<__Key,__PathItemData>::CEAS_ADIB(CEASRoutingUnit<__Key,__PathItemData>* routingUnit)
  : CEASRoutingUnitIB<__Key,CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>*,__PathItemData>(routingUnit)
{
    D_=static_cast<unsigned int>(atoi(CEASNetworkLayerUnit::getTclConfig("D")));
    rho_ = atof(CEASNetworkLayerUnit::getTclConfig("rho"));
    rhoReductionFactor_ = atof(
      CEASNetworkLayerUnit::getTclConfig("rho-reduction-factor"));
    temperatureType_ = static_cast<CEASTemperature_t>(atoi(
      CEASNetworkLayerUnit::getTclConfig("temperature-approx")));
}

//! Create and add a new entry in the ADIB for the given key
/*! If there is already an entry for this key in the ADIB, it is deleted. 
   \returns the entry created */
template<class __Key,class __PathItemData> CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>*
CEAS_ADIB<__Key,__PathItemData>::addEntry(const __Key& key)
{
    // create a new entry
    CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>* entry = 0;
    switch (temperatureType_) {
      case CEAS_TEMPERATURE_HYPERBOLIC_APPROX:
        entry = (CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>*)(
          new CEAS_ADIBEntry<__Key,CEASTemperatureHyp*,__PathItemData>(
            routingUnit_));
        break;
      case CEAS_TEMPERATURE_LINEAR_APPROX:
        entry = (CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>*)(
          new CEAS_ADIBEntry<__Key,CEASTemperatureLin*,__PathItemData>(
            routingUnit_));
        break;
      default:
        fprintf(stderr,"CEAS_ADIB::%s: Invalid temperature type\n",
                __FUNCTION__);
        abort();
    }
    // add the new entry in the ADIB
    pair<typename map<__Key,CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>*>::iterator,
      bool> p = base_.insert(map_make_pair<__Key,CEAS_ADIBEntry<__Key,
                             CEASTemperature*,__PathItemData>*>(key,entry));
    if (!p.second)
      { // there is already an entry with this key, replace it
        CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>* oldEntry =
          static_cast<CEAS_ADIBEntry<__Key,CEASTemperature*,__PathItemData>*>(
            (p.first)->second);
        delete oldEntry;
        base_.erase(p.first);
        p = base_.insert(map_make_pair<__Key,CEAS_ADIBEntry<__Key,
                         CEASTemperature*,__PathItemData>*>(key,entry));
        assert(p.second);
      }
      return entry;
}
#endif

/*! ADIB (At Destination Information Base) Entry*/
#ifndef stand_alone 
template<class _Key,class _CEASTemperature,class _PathItemData>
class CEAS_ADIBEntry<_Key,_CEASTemperature*,_PathItemData> {
  public:
    CEAS_ADIBEntry(CEASRoutingUnit<_Key,_PathItemData>* routingUnit);
#else
template<class _CEASTemperature>
class CEAS_ADIBEntry<_CEASTemperature*> {
  public:
    CEAS_ADIBEntry(double rho,double rhoReductionFactor,unsigned int D);
#endif
    ~CEAS_ADIBEntry();
    inline _CEASTemperature* totalTemperature() {return totalTemperature_;}
    inline _CEASTemperature* eliteTemperature() {return eliteTemperature_;}
    inline double eliteSelectionLevel()
      {return (double)(-log(rho_) * totalTemperature_->gamma());} 
    inline double updateTotalTemperature(double L,double beta);
    inline double updateEliteTemperature(double L,double beta);
    void updateRho(double L);
    inline double rho() const {return rho_;}
  protected:
    /*! temperature based on the costs of paths found by all the forward ants
        received */
    _CEASTemperature* totalTemperature_;
    /*! temperature based on the costs of paths below the elite selection level
        at the arrival of the forward ant */
    _CEASTemperature* eliteTemperature_;
    double bestCost_; //!< best cost observed so far
    unsigned int receivedSinceBest_;//!< number of ants received since last best
    double rho_; //!< search focus
    double rhoReductionFactor_; //!< rho reduction factor
    unsigned int D_; //!< control parameter used for rho reduction (#iterations)
#ifndef stand_alone
    CEASRoutingUnit<_Key,_PathItemData>* routingUnit_;
#endif
};

/*! Constructor */
#ifndef stand_alone
template<class _Key,class _CEASTemperature,class _PathItemData>
CEAS_ADIBEntry<_Key,_CEASTemperature*,_PathItemData>::
CEAS_ADIBEntry(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
  : bestCost_(MAXDOUBLE),receivedSinceBest_(0),routingUnit_(routingUnit)
{
    totalTemperature_ = new _CEASTemperature();
    if (routingUnit->eliteSelection())
      eliteTemperature_ = new _CEASTemperature();
    else eliteTemperature_ = 0;
    // get initial parameter values from the adib
    CEAS_ADIB<_Key,_PathItemData>* adib = 
      static_cast<CEAS_ADIB<_Key,_PathItemData>*>(&(routingUnit->adib())); 
    D_ = adib->D();
    rho_ = adib->rho();
    rhoReductionFactor_ = adib->rhoReductionFactor();
}
#else
template<class _CEASTemperature>
CEAS_ADIBEntry<_CEASTemperature*>::CEAS_ADIBEntry(double rho,
                                                  double rhoReductionFactor,
                                                  unsigned int D)
  : bestCost_(MAXDOUBLE),receivedSinceBest_(0),rho_(rho),
    rhoReductionFactor_(rhoReductionFactor),D_(D)
{
    totalTemperature_ = new _CEASTemperature();
    eliteTemperature_ = new _CEASTemperature();
}
#endif


/*! Destructor */
#ifndef stand_alone
template<class _Key,class _CEASTemperature,class _PathItemData>
CEAS_ADIBEntry<_Key,_CEASTemperature*,_PathItemData>::
#else
template<class _CEASTemperature>
CEAS_ADIBEntry<_CEASTemperature*>::
#endif
~CEAS_ADIBEntry()
{
    if (totalTemperature_) delete totalTemperature_;
    if (eliteTemperature_) delete eliteTemperature_;
}

//! Update the total temperature
/*! \returns the new total gamma value */
#ifndef stand_alone
template<class _Key,class _CEASTemperature,class _PathItemData> double
CEAS_ADIBEntry<_Key,_CEASTemperature*,_PathItemData>::
#else
template<class _CEASTemperature> double
CEAS_ADIBEntry<_CEASTemperature*>::
#endif
updateTotalTemperature(double L,double beta)
{
#ifndef stand_alone
    if (routingUnit_->logger())
      routingUnit_->logger()->sample(routingUnit_,CEAS_LOG_GAMMA_CALC_COUNT);
#endif
    return totalTemperature_->update(L,beta,rho_);
}

//! Update the elite temperature
/*! \returns the new elite gamma value */
#ifndef stand_alone
template<class _Key,class _CEASTemperature,class _PathItemData> double
CEAS_ADIBEntry<_Key,_CEASTemperature*,_PathItemData>::
#else
template<class _CEASTemperature> double
CEAS_ADIBEntry<_CEASTemperature*>::
#endif
updateEliteTemperature(double L,double beta)
{
    assert(eliteTemperature_);
#ifndef stand_alone
    if (routingUnit_->logger())
      routingUnit_->logger()->sample(routingUnit_,CEAS_LOG_GAMMA_CALC_COUNT);
#endif
    return eliteTemperature_->update(L,beta,rho_);
}

//! Adjust rho if no new best path has been found by the last D agents
#ifndef stand_alone
template<class _Key,class _CEASTemperature,class _PathItemData> void
CEAS_ADIBEntry<_Key,_CEASTemperature*,_PathItemData>::
#else
template<class _CEASTemperature> void
CEAS_ADIBEntry<_CEASTemperature*>::
#endif
updateRho(double L)
{
    if (L < bestCost_)
      { // new best cost
        bestCost_ = L;
        receivedSinceBest_ = 0; // reset
#ifndef stand_alone
        if (routingUnit_->logger())
          routingUnit_->logger()->sample(routingUnit_,CEAS_LOG_COST_BEST);
#endif
      }
    else
      { // no new best
        receivedSinceBest_++; // increment
        if (receivedSinceBest_ > D_)
          { // adjust rho reduction factor
            rho_ *= rhoReductionFactor_;
            receivedSinceBest_ = 0;
#ifndef stand_alone
            if (routingUnit_->logger())
              routingUnit_->logger()->sample(routingUnit_,CEAS_LOG_RHO);
#endif
          }
      }
}
    
#endif    
