/*
 * Copyright (c) 2008 Q2S NTNU, Trondheim, Norway
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Port of the Swarm module developed by O. Wittner for ns-2.30.
 * Author: Laurent Paquereau <laurent.paquereau@q2s.ntnu.no>
 *
 * $Header$
 */

#ifndef _ceas_logger_h
#define _ceas_logger_h

// Log flags
#define CEAS_LOG_COST                        (unsigned int) 0x00001
#define CEAS_LOG_COST_ALL                    (unsigned int) 0x00002
#define CEAS_LOG_COST_BEST                   (unsigned int) 0x00004
#define CEAS_LOG_PATH                        (unsigned int) 0x00008
#define CEAS_LOG_PATH_ALL                    (unsigned int) 0x00010
#define CEAS_LOG_DELAY                       (unsigned int) 0x00020
#define CEAS_LOG_ELITE_GAMMA                 (unsigned int) 0x00040
#define CEAS_LOG_TOTAL_GAMMA                 (unsigned int) 0x00080
#define CEAS_LOG_GAMMA_DIFF                  (unsigned int) 0x00100
#define CEAS_LOG_GAMMA_REL                   (unsigned int) 0x00200
#define CEAS_LOG_ELITE_SELECTION_LEVEL       (unsigned int) 0x00400
#define CEAS_LOG_RHO                         (unsigned int) 0x00800
#define CEAS_LOG_BACKTRACKCOUNT              (unsigned int) 0x01000
#define CEAS_LOG_FORWARDCOUNT                (unsigned int) 0x02000
#define CEAS_LOG_PHEROMONES                  (unsigned int) 0x04000
#define CEAS_LOG_GAMMA_CALC_COUNT            (unsigned int) 0x08000
#define CEAS_LOG_TOTAL_TEMPERATURE_VARS      (unsigned int) 0x10000

#include <set>
#include <math.h>
#include <logger.h>
#include <neighbour-information-base.h>
#include "ceas-network-layer-unit.h"
#include "ceas-packet.h"
#include "ceas-routing-unit.h"
#include "ceas-utils.h"

// Forward declaration
template<class _Key,class _PathItemData> class CEASRoutingUnit;

//! Logger class for CEAS
template<class _Key = int32_t,class _PathItemData = CEASPathItemData>
class CEASLogger : public Logger {
  public:
    CEASLogger();
    virtual void sample(void*, unsigned int);
    virtual void samplePhero(void*);
    static CEASLogger<_Key,_PathItemData>* instance()
    {
        if (!instance_)
          {
            instance_ = new CEASLogger<_Key,_PathItemData>();
            instance_->init();
          }
        return instance_;
    }
    virtual void init();
  protected:
    virtual const char* prefix() const {return "ceas";}
    void log_printf(FILE*& file,
                    _Key& k,
                    const char* postfix,
                    const char *fmt,
                    ...);
    void log_printf(FILE*& file,
                    char* addr,
                    const char* postfix,
                    const char *fmt,
                    ...);
    static CEASLogger<_Key,_PathItemData>* instance_;
    double geoWeight_; //!< weight for geometrically weighted average
    double samplePeriod_; //!< min period between samples to be logged
    //! logging flag mask
    /*! configured using CEAS-config -log-cost|path|...*/
    u_int32_t mask_;
    //! set of keys to be logged, empty = all
    /*! configured  using CEAS-config -log-only */
    set<_Key> logOnly_;
    inline const char* getTclConfig(const char* param);
    //! abscissa (1 column), default is time
    virtual double x(void*) const {return NOW;}
    /*! key to string conversion */
    char* key2str(_Key&);
    /*! string to key conversion */
    void str2key(_Key&,char*);

    map_file<_Key> costFile_; //!< Log file handle
    map_double<_Key> costLast_; //!< Time when last sample was logged
    void logCost(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<_Key> costAllFile_; //!< Log file handle
    map_double<_Key> costAllLast_; //!< Time when last sample was logged
    void logCostAll(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<_Key> costBestFile_;
    void logCostBest(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<_Key> pathFile_; //!< Log file handle
    map_double<_Key> pathLast_; //!< Time when last sample was logged
    void logPath(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<_Key> pathAllFile_; //!< Log file handle
    map_double<_Key> pathAllLast_; //!< Time when last sample was logged
    void logPathAll(CEASRoutingUnit<_Key,_PathItemData>*);

    template<class T> static char* path2str(list<pair<nsaddr_t,T> >& path);

    map_double<_Key> delay_avg_;
    map_file<_Key> delayFile_; //!< Log file handle
    void logDelay(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<_Key> eliteGammaFile_; //!< Log file handle
    map_double<_Key> eliteGammaLast_;//!<Time when last sample was logged
    void logEliteGamma(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<_Key> totalGammaFile_; //!< Log file handle
    map_double<_Key> totalGammaLast_; //!< Time when last sample was logged
    void logTotalGamma(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<_Key> totalTemperatureVarsFile_; //!< Log file handle
    map_double<_Key> totalTemperatureVarsLast_; //!< Time when last sample was logged
    void logTotalTemperatureVars(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<_Key> gammaDiffFile_; //!< Log file handle
    map_double<_Key> gammaDiffLast_; //!< Time when last sample was logged
    void logGammaDiff(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<_Key> gammaRelFile_; //!< Log file handle
    map_double<_Key> gammaRelLast_; //!< Time when last sample was logged
    void logGammaRel(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<_Key> eliteSelectionLevelFile_; //!< Log file handle
    //! Time when last sample was logged
    map_double<_Key> eliteSelectionLevelLast_;
    void logEliteSelectionLevel(CEASRoutingUnit<_Key,_PathItemData>*);

    map_uint<_Key> gammaCalcCount_; 
    map_file<_Key> gammaCalcCountFile_; //!< Log file handle
    map_double<_Key> gammaCalcCountLast_; //!< Time when last sample was logged
    void logGammaCalcCount(CEASRoutingUnit<_Key,_PathItemData>*);

    map_uint<_Key> backtrackCount_; 
    map_file<_Key> backtrackCountFile_; //!< Log file handle
    map_double<_Key> backtrackCountLast_; //!< Time when last sample was logged
    void logBacktrackCount(CEASRoutingUnit<_Key,_PathItemData>*);

    map_uint<_Key> forwardCount_; 
    map_file<_Key> forwardCountFile_; //!< Log file handle
    map_double<_Key> forwardCountLast_; //!< Time when last sample was logged
    void logForwardCount(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<_Key> rhoFile_; //!< Log file handle
    map_double<_Key> rhoLast_; //!< Time when last sample was logged
    void logRho(CEASRoutingUnit<_Key,_PathItemData>*);

    map_file<nsaddr_t> pheromonesFile_; //!< Log file handle
    map_double<nsaddr_t> pheromonesLast_;//!< Time when last sample was logged
    //! Last phero level sampled for a pheroId and a neighbour key
    map<nsaddr_t,map<int32_t,map<NeighbourKey,double,NeighbourKey> > > phero_last_level_;
    //! Gets the pheroId from the key
    int32_t pheroId(_Key&);
    /*! \returns a string version of the given pheroId*/
    char* pheroId2str(CEASLogger<_Key,_PathItemData>*,int32_t);
    //! set of pheroId to log, empty = all
    /*! configured using CEAS-config -log-pheromones-id */
    set<int32_t> logPheroId_;
    //! set of nodes at which to log phero distr, empty = all
    /*! configured using CEAS-config -log-pheromones-at */
    set<nsaddr_t> logPheroAt_;
    void logPhero(CEASRoutingUnit<_Key,_PathItemData>*);
};

// Static variable
template<class _Key,class _PathItemData> CEASLogger<_Key,_PathItemData>* CEASLogger<_Key,_PathItemData>::instance_;

/*! \returns value (const char*) of param set in OTcl*/
template<class _Key,class _PathItemData> const char*
CEASLogger<_Key,_PathItemData>::getTclConfig(const char* param)
{
     Tcl& tcl = Tcl::instance();
     tcl.evalf("Logger/CEAS set config_(%s)",param);
     return (tcl.result());
}

//! Constructor
template<class _Key,class _PathItemData>
CEASLogger<_Key,_PathItemData>::CEASLogger(): Logger(), mask_(0)
{
}

//! Parse and initialized Tcl configured parameters
template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::init()
{
    // Set sample frequency
    double freq = atof(getTclConfig("sample-freq"));
    samplePeriod_ = (freq > 0 ? 1/freq : 0);
    
    // Set geometric weight to make sample no. "bin-size" 
    // back in time weighted by 0.3
    double bin_size = atof(getTclConfig("bin-size"));
    geoWeight_ = bin_size > 0 ? exp(log(0.3)/bin_size) : 0 ;
    
    // Set log mask
    if (atoi(getTclConfig("cost")))
      mask_ |= CEAS_LOG_COST;
    if (atoi(getTclConfig("cost-all")))
      mask_ |= CEAS_LOG_COST_ALL;
    if (atoi(getTclConfig("cost-best")))
      mask_ |= CEAS_LOG_COST_BEST;
    if (atoi(getTclConfig("path")))
      mask_ |= CEAS_LOG_PATH;
    if (atoi(getTclConfig("path-all")))
      mask_ |= CEAS_LOG_PATH_ALL;
    if (atoi(getTclConfig("delay")))
      mask_ |= CEAS_LOG_DELAY;
    if (atoi(getTclConfig("elite-gamma")))
      mask_ |= CEAS_LOG_ELITE_GAMMA;
    if (atoi(getTclConfig("total-gamma")))
      mask_ |= CEAS_LOG_TOTAL_GAMMA;
    if (atoi(getTclConfig("gamma-diff")))
      mask_ |= CEAS_LOG_GAMMA_DIFF;
    if (atoi(getTclConfig("gamma-rel")))
      mask_ |= CEAS_LOG_GAMMA_REL;
    if (atoi(getTclConfig("elite-selection-level")))
      mask_ |= CEAS_LOG_ELITE_SELECTION_LEVEL;
    if (atoi(getTclConfig("rho")))
      mask_ |= CEAS_LOG_RHO;
    if (atoi(getTclConfig("backtrackcount")))
      mask_ |= CEAS_LOG_BACKTRACKCOUNT;
    if (atoi(getTclConfig("forwardcount")))
      mask_ |= CEAS_LOG_FORWARDCOUNT;
    if (atoi(getTclConfig("pheromones")))
      mask_ |= CEAS_LOG_PHEROMONES;
    if (atoi(getTclConfig("gamma-calc-count")))
      mask_ |= CEAS_LOG_GAMMA_CALC_COUNT;
    if (atoi(getTclConfig("total-temperature-vars")))
      mask_ |= CEAS_LOG_TOTAL_TEMPERATURE_VARS;
    
    if (mask_ & CEAS_LOG_PHEROMONES)
      {
        const char* str = getTclConfig("pheromones-id");
        // a comma separated list of phero id to log is expected
        // convert list to set of integers
        char id_str[strlen(str)+1];
        strncpy(id_str,str,strlen(str)+1);
        char* next = strtok(id_str,",");
        while (next)
          {
            int32_t id;
            if (sscanf(next,"%d",&id)==1) logPheroId_.insert(id);
            next = strtok(NULL,",");
          }
        str = getTclConfig("pheromones-at");
        // a comma-list of nodes/node addresses where to log phero distr.
        char node_str[strlen(str)+1];
        strncpy(node_str,str,strlen(str)+1);
        next = strtok(node_str,",");
        while (next)
          {
            Node* node = dynamic_cast<Node*>(TclObject::lookup(next));
            if (node)
              // node object
              logPheroAt_.insert(node->address());
            else
              // node address
              logPheroAt_.insert(Address::instance().str2addr(next));
            next = strtok(NULL,",");
          }
      }
    
    const char* str = getTclConfig("only");
    char key_str[strlen(str)+1];
    strncpy(key_str,str,strlen(str)+1);
    char* next = strtok(key_str,",");
    while (next)
      {
        _Key k;
        str2key(k,next);
        logOnly_.insert(k);  
        next = strtok(NULL,",");
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::sample(void* data,unsigned int flags)
{
    flags &= mask_;
    CEASRoutingUnit<_Key,_PathItemData>* routingUnit =
      static_cast<CEASRoutingUnit<_Key,_PathItemData>*>(data);
    _Key k = routingUnit->key();
    if (!flags || !(logOnly_.empty() || logOnly_.find(k)!=logOnly_.end()))
      // do not log
      return;
    if (flags & CEAS_LOG_COST) logCost(routingUnit);
    if (flags & CEAS_LOG_COST_ALL) logCostAll(routingUnit);
    if (flags & CEAS_LOG_COST_BEST) logCostBest(routingUnit);
    if (flags & CEAS_LOG_PATH) logPath(routingUnit);
    if (flags & CEAS_LOG_PATH_ALL) logPathAll(routingUnit);
    if (flags & CEAS_LOG_DELAY) logDelay(routingUnit);
    if (flags & CEAS_LOG_ELITE_GAMMA) logEliteGamma(routingUnit);
    if (flags & CEAS_LOG_TOTAL_GAMMA) logTotalGamma(routingUnit);
    if (flags & CEAS_LOG_GAMMA_DIFF) logGammaDiff(routingUnit);
    if (flags & CEAS_LOG_GAMMA_REL) logGammaRel(routingUnit);
    if (flags & CEAS_LOG_ELITE_SELECTION_LEVEL)
      logEliteSelectionLevel(routingUnit);
    if (flags & CEAS_LOG_GAMMA_CALC_COUNT) logGammaCalcCount(routingUnit);
    if (flags & CEAS_LOG_BACKTRACKCOUNT) logBacktrackCount(routingUnit);
    if (flags & CEAS_LOG_FORWARDCOUNT) logForwardCount(routingUnit);
    if (flags & CEAS_LOG_RHO) logRho(routingUnit);
    if (flags & CEAS_LOG_TOTAL_TEMPERATURE_VARS) 
      logTotalTemperatureVars(routingUnit);
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::samplePhero(void* data)
{
    unsigned int flags = CEAS_LOG_PHEROMONES;
    flags &= mask_;
    CEASRoutingUnit<_Key,_PathItemData>* routingUnit =
      static_cast<CEASRoutingUnit<_Key,_PathItemData>*>(data);
    int32_t id = pheroId(routingUnit->key());
    nsaddr_t at = routingUnit->local_address();

    if (!flags ||
        !(logPheroAt_.empty() || logPheroAt_.find(at)!=logPheroAt_.end()) ||
        !(logPheroId_.empty() || logPheroId_.find(id)!=logPheroId_.end()))
      // do not log
      return;
    logPhero(routingUnit);
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logCost(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    CEASData<_PathItemData>* ant = routingUnit->ant();
    if ( NOW - costLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(costFile_[k],
                   k,
                   "cost",
                   "%g %g %d\n",
                   x(routingUnit),ant->L(),ant->behaviour());
        costLast_[k] = NOW;
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logCostAll(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    CEASData<_PathItemData>* ant = routingUnit->ant();
    if ( NOW - costAllLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(costAllFile_[k],
                   k,
                   "cost_all",
                   "%g %g %d\n",
                   x(routingUnit),ant->L(),ant->behaviour());
        costAllLast_[k] += samplePeriod_;
    }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logCostBest(
  CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    CEASData<_PathItemData>* ant = routingUnit->ant();
    log_printf(costBestFile_[k],
               k,
               "cost_best",
               "%g %g %d\n",
               x(routingUnit),ant->L(),ant->behaviour());
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logPath(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    CEASData<_PathItemData>* ant = routingUnit->ant();
    char* pathstr = path2str(ant->path());
    log_printf(pathFile_[k],
               k,
               "path",
               "%g [ %s ] %g %d\n",
               x(routingUnit),pathstr,ant->L(),ant->behaviour());
    pathLast_[k] = NOW;
    delete pathstr;
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logPathAll(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    CEASData<_PathItemData>* ant = routingUnit->ant();
    char* pathstr = path2str(ant->path());
    log_printf(pathAllFile_[k],
               k,
               "path_all",
               "%g [ %s ] %g %d\n",
               x(routingUnit),pathstr,ant->L(),ant->behaviour());
    pathAllLast_[k] = NOW;
    delete pathstr;
}



template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logDelay(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logEliteGamma(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    if ( NOW - eliteGammaLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(eliteGammaFile_[k],
                   k,
                   "gamma_elite",
                   "%g %g\n",
                   x(routingUnit),routingUnit->adib()[k]->eliteTemperature()->gamma());
        eliteGammaLast_[k] = NOW;
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logTotalGamma(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    if ( NOW - totalGammaLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(totalGammaFile_[k],
                   k,
                   "gamma_all",
                   "%g %g\n",
                   x(routingUnit),routingUnit->adib()[k]->totalTemperature()->gamma());
        totalGammaLast_[k] = NOW;
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logTotalTemperatureVars(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    if ( NOW - totalTemperatureVarsLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(totalTemperatureVarsFile_[k],
                   k,
                   "temperature_vars_all",
                   "%g %g %g\n",
                   x(routingUnit),
                   routingUnit->adib()[k]->totalTemperature()->a(),
                   routingUnit->adib()[k]->totalTemperature()->b());
        totalTemperatureVarsLast_[k] = NOW;
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logGammaDiff(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    if ( NOW - gammaDiffLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(gammaDiffFile_[k],
                   k,
                   "gamma_diff",
                   "%g %g\n",
                   x(routingUnit),routingUnit->adib()[k]->totalTemperature()->gamma()-
                   routingUnit->adib()[k]->eliteTemperature()->gamma());
        gammaDiffLast_[k] = NOW;
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logGammaRel(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    if ( NOW - gammaRelLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(gammaRelFile_[k],
                   k,
                   "gamma_rel","%g %g\n",
                   x(routingUnit),routingUnit->adib()[k]->totalTemperature()->gamma()/
                   routingUnit->adib()[k]->eliteTemperature()->gamma());
        gammaRelLast_[k] = NOW;
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logEliteSelectionLevel(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    if ( NOW - eliteSelectionLevelLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(eliteSelectionLevelFile_[k],
                   k,
                   "elitelimit",
                   "%g %g\n",
                   x(routingUnit),routingUnit->adib()[k]->eliteSelectionLevel());
        eliteSelectionLevelLast_[k] = NOW;
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logGammaCalcCount(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    gammaCalcCount_[k]++;
    if ( NOW - gammaCalcCountLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(gammaCalcCountFile_[k],
                   k,
                   "gamma_calc_count",
                   "%g %d\n",
                   x(routingUnit),
                   gammaCalcCount_[k]);
        gammaCalcCountLast_[k] = NOW;
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logBacktrackCount(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    backtrackCount_[k]++;
    if ( NOW - backtrackCountLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(backtrackCountFile_[k],
                   k,
                   "backtrackcount",
                   "%g %d\n",
                   x(routingUnit),
                   backtrackCount_[k]);
        backtrackCountLast_[k] = NOW;
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logForwardCount(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    forwardCount_[k]++;
    if ( NOW - forwardCountLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(forwardCountFile_[k],
                   k,
                   "forwardcount",
                   "%g %d\n",
                   x(routingUnit),
                   forwardCount_[k]);
        forwardCountLast_[k] = NOW;
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logRho(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    _Key k = routingUnit->key();
    if ( NOW - rhoLast_[k] >= samplePeriod_ )
      { // Log sample
        log_printf(rhoFile_[k],
                   k,
                   "rho","%g %g\n",
                   x(routingUnit),routingUnit->adib()[k]->rho());
        rhoLast_[k] = NOW;
      }
}

template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::logPhero(CEASRoutingUnit<_Key,_PathItemData>* routingUnit)
{
    int32_t id = pheroId(routingUnit->key());
    nsaddr_t at_node = routingUnit->local_address();
    char* at_node_str = addr2str(at_node);
    Packet* p = routingUnit->p();
    hdr_ip*  ih = hdr_ip::access(p);
    vector<pair<NeighbourEntry*,double> > phero_dstr;
    CEASNetworkLayerUnit* networkLayerUnit = routingUnit->networkLayerUnit();
    networkLayerUnit->getAllNeighboursPheroLevel(routingUnit->species(),
                                                 ih->daddr(),
                                                 ih->saddr(),
                                                 phero_dstr);
    vector<pair<NeighbourEntry*,double> >::iterator i = phero_dstr.begin();
    for(; i != phero_dstr.end(); i++)
      phero_last_level_[at_node][id][i->first->key()] = i->second;
    if ( NOW - pheromonesLast_[at_node] >= samplePeriod_ )
      {
        map<int,map<NeighbourKey,double,NeighbourKey> >::iterator j =
          phero_last_level_[at_node].begin();
        for(;j !=  phero_last_level_[at_node].end(); j++)
          {
            // log time and phero id
            char* phero_id_str = pheroId2str(this,j->first);
            log_printf(pheromonesFile_[at_node],
                       at_node_str,
                       "phero",
                       "%g %s ",
                       x(routingUnit),
                       phero_id_str);
            delete phero_id_str;
            // for each neighbour log neighbour and phero level
            map<NeighbourKey,double,NeighbourKey>::iterator k=j->second.begin();
            for(; k != j->second.end(); k++)
              {
                char* addr = addr2str(k->first.getAddress());
                log_printf(pheromonesFile_[at_node],
                           at_node_str,
                           "phero",
                           "[ %s %d %d ] %g ",
                           addr,
                           k->first.getInterfaceId(),
                           k->first.getChannelIndex(),
                           k->second);
                delete addr;
              }
            // add newline
            log_printf(pheromonesFile_[at_node],at_node_str,"phero","\n");
          }
        pheromonesLast_[at_node] = NOW;
      }
    delete at_node_str;
}

//! Store formatted string to log file indexed by key
template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::log_printf(FILE*& file,
                             _Key& key,
                             const char* postfix,
                             const char *fmt,
                             ...)
{
    if (! file)
      {
        // Prepare filename
        char filename[400];
        char* keystr = key2str(key);
        sprintf(filename,"%s_%s_%s",prefix(),keystr,postfix);
        delete[] keystr;
        // Open file
        if(!(file = fopen(filename, "w")))
          {
            printf("CEASLogger<_Key,_PathItemData>::log_printf: Cannot open file %s\n",
                   filename);
            return;
          }
      }
    // Output string to file
    va_list ap;
    va_start(ap, fmt);
    vfprintf(file, fmt, ap);
    va_end(ap);
    fflush(file);
}

//! Store formatted string to log file indexed by key
template<class _Key,class _PathItemData> void
CEASLogger<_Key,_PathItemData>::log_printf(FILE*& file,
                             char* addr,
                             const char* postfix,
                             const char *fmt,
                             ...)
{
    if (! file)
      {
        // Prepare filename
        char filename[400];
        sprintf(filename,"%s_%s_%s",prefix(),addr,postfix);
        // Open file
        if(!(file = fopen(filename,"w")))
          {
            printf("CEASLogger<_Key,_PathItemData>::log_printf: Cannot open file %s\n",
                   filename);
            return;
          }
      }
    // Output string to file
    va_list ap;
    va_start(ap, fmt);
    vfprintf(file, fmt, ap);
    va_end(ap);
    fflush(file);
}

/*! \returns a string version of the path (list of node addresses) */
template<class _Key,class _PathItemData> template<class T> char*
CEASLogger<_Key,_PathItemData>::path2str(list<pair<nsaddr_t,T> >& path)
{
    assert(!path.empty());
    char* pathstr = 0;
    char str[path.size()*15+1];
    str[0]='\0';
    typename list<pair<nsaddr_t,T> >::iterator i = path.begin();
    for (;;)
      {
        char* addr = addr2str(i->first);
        strcat(str,addr);
        delete addr;
        i++;
        if (i!=path.end()) strcat(str," ");
        else break;
      }
    pathstr = new char[strlen(str)+1];
    strcpy(pathstr,str);
    return pathstr;
}

#endif
