/*
 * Copyright (c) 2008 Q2S NTNU, Trondheim, Norway
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Port of the Swarm module developed by O. Wittner for ns-2.30.
 * Author: Laurent Paquereau <laurent.paquereau@q2s.ntnu.no>
 *
 * $Header$
 */

#include "address.h"
#include "drop.h"
#include "forwarding-unit.h"
#include "logger.h"
#include "node.h"
#include "routing-unit.h"
#include "ceas-logger.h"
#include "ceas-network-layer-unit.h"
#include "ceas-packet-generator.h"
#include "ceas-wired-neighbour.h"

// OTcl linkage
static class CEASNetworkLayerUnitClass : public TclClass {
public:
    CEASNetworkLayerUnitClass()
      : TclClass("NetworkLayerUnit/CEAS") {}
    TclObject* create(int argc, const char*const* argv)
    {
        fprintf(stderr, "CEASNetworkLayerUnitClass::%s: Generic class. "\
                "Use subclasses.\n",__FUNCTION__);
        abort();
    }
} class_CEASNetworkLayerUnit;

//! Constructor
CEASNetworkLayerUnit::CEASNetworkLayerUnit(Node* node)
  : NetworkLayerUnit(node),randomgen_(0)
{
    Tcl& tcl = Tcl::instance();
    // Create a pheromone table
    tcl.evalf("new NetworkInformationBase/RIB/CEASPheroTable %s",
              node->name());
    rib_ = dynamic_cast<CEASPheroTable*>(TclObject::lookup(tcl.result()));
    rib_->setNetworkLayerUnit(this);
    // Create a neighbour information base table
    tcl.evalf("new NeighbourInformationBase/CEAS %s",node->name());
    neighbourInformationBase_ = dynamic_cast<CEASNeighbourInformationBase*>(
      TclObject::lookup(tcl.result()));
}

//! Destructor
CEASNetworkLayerUnit::~CEASNetworkLayerUnit()
{
}

//! Process commands fired at TCL-Level
/*! \returns TCL_OK if the command has been completed, TCL_ERROR otherwise. */
int
CEASNetworkLayerUnit::command(int argc, const char*const* argv)
{
    Tcl& tcl = Tcl::instance();
    if (argc == 2)
      {
        if (strcmp(argv[1],"attach-logger") == 0)
          { // the logger must have been created by a subclass
            if (!logger_) logger_ = CEASLogger<>::instance();
            // Set ref to logger for the forwarding and routing units
            if (forwardingUnit_) forwardingUnit_->setLogger(logger_);
            if (routingUnit_) routingUnit_->setLogger(logger_);
            // Set ref to log in all generators
            list<RoutingPacketGenerator*>::iterator i = gList_.begin();
            for(; i != gList_.end(); i++) (*i)->setLogger(logger_);
            return TCL_OK;
          }
      }
    else if (argc == 3)
      { // no arg in addition to the command
        if (strcmp(argv[1],"use-rng") == 0)
          { // set the random generator
            randomgen_ = dynamic_cast<RNG*>(TclObject::lookup(argv[2]));
            if (!randomgen_)
              {
                tcl.add_errorf("CEASNetworkLayerUnit::%s (use-rng): "\
                               "Lookup failed (%s)",__FUNCTION__,argv[2]);
                return TCL_ERROR;
              }
            return TCL_OK;
          }
      }
     return (NetworkLayerUnit::command(argc, argv));
}

void
CEASNetworkLayerUnit::addInterface(NetworkInterface2* iface)
{
    NetworkLayerUnit::addInterface(iface);
    PointToPointInterface* pp = dynamic_cast<PointToPointInterface*>(iface);
    if (pp && pp->isOutputInterface())
      neighbourInformationBase_->addEntry(new CEASWiredNeighbour(pp));
}

void
CEASNetworkLayerUnit::processPacket(Packet* p,Handler* h){
    hdr_cmn* ch = hdr_cmn::access(p);
    hdr_ceas* ceash = hdr_ceas::access(p);
    // classify packets
    if (ch->ptype()==packetType() && ceash->type()==ceasType())
      {
        assert(routingUnit_);
        routingUnit_->recv(p,h);
      }
    else
      {
        assert(forwardingUnit_);
        forwardingUnit_->recv(p,h);
      }
}

/*! Send p to the given neighbour */
void
CEASNetworkLayerUnit::sendDown(Packet* p,Handler* h,NeighbourEntry* neighbour)
{
    assert(neighbour);
    hdr_cmn* ch = hdr_cmn::access(p);
    ch->direction() = hdr_cmn::DOWN;
    ch->prev_hop() = node_->address();
    ch->next_hop() = neighbour->getAddress();
    ch->iface() = neighbour->getInterface()->getId();
    ch->channel_index() = neighbour->getChannel() ?
      neighbour->getChannel()->index() : UNKN_CHANNEL;
    NetworkLayerUnit::sendDown(p,h,neighbour->getInterface()->getId());
}

/*! Select neighbour by applying pheromone distribution.
    If the sum of pherolevels is less than PHERO_VERY_SMALL, use a uniform
    distribution instead.
    \param neighbours should be a list of neighbours with the associated
    accumulated pheromone level so that neighbours.back().second is the sum of
    all pherolevels.
 */
NeighbourEntry*
CEASNetworkLayerUnit::selectRandomNeighbour(
  vector<pair<NeighbourEntry*,double> >& neighbours)
{
    assert(randomgen_);
    assert(!neighbours.empty());
    if (neighbours.size() == 1)
      return neighbours.begin()->first;
    double pherolevel_sum = neighbours.back().second;
    if (pherolevel_sum < PHERO_VERY_SMALL) 
      { // No pheromones found. Make a uniform dist.
        long n = randomgen_->rand_int(1,neighbours.size());
        return neighbours[n-1].first;
      }
    // Apply "roulette wheel" method base on pheromone
    // distribution to select next node
    double u = randomgen_->uniform_double();
    vector<pair<NeighbourEntry*,double> >::iterator i = neighbours.begin();
    for (; i != neighbours.end(); i++)
      {
        if (u <= (i->second / pherolevel_sum))
          // Select neighbour to be next
          return i->first;
      }
    return 0; // should never happen
}

/*! Build list of all neighbours. Do not consider the pheromone level
    associated with each of the neighbours. Do not consider either whether the 
    neighbour is reachable or not*/
void
CEASNetworkLayerUnit::getAllNeighbours(
  vector<pair<NeighbourEntry*,double> >& neighbours)
{
    neighbourMap_t* neighbourhood = neighbourInformationBase_->getBase();
    neighbourMap_t::iterator i = neighbourhood->begin();
    for (;  i != neighbourhood->end(); i++)
      { // for each neighbour entry
        NeighbourEntry* neighbour = static_cast<NeighbourEntry*>(i->second);
        neighbours.push_back(pair<NeighbourEntry*,double>(neighbour,0));
      }
}

/*! Build the list of (neighbour,pheromone level towards the neighbour) for the 
    given saddr, daddr and species */
void
CEASNetworkLayerUnit::getAllNeighboursPheroLevel(
  int32_t species,
  nsaddr_t src,
  nsaddr_t dst,
  vector<pair<NeighbourEntry*,double> >& neighbours)
{
    neighbourMap_t* neighbourhood = neighbourInformationBase_->getBase();
    neighbourMap_t::iterator i = neighbourhood->begin();
    for (;  i != neighbourhood->end(); i++)
      { // for each neighbour entry
        NeighbourEntry* neighbour = dynamic_cast<NeighbourEntry*>(i->second);
        // get the entry in the pherotable for this neighbour, if any
        CEASPheroTableEntry* entry = pherotable()->findEntry(neighbour,
                                                             species,
                                                             src,
                                                             dst);
        // store neigbour and pherolevel towards neigbour
        neighbours.push_back(
          pair<NeighbourEntry*,double>(neighbour,
                                       entry ? entry->getPheroLevel():0));
      }
}

/*! Build the list of (reachable neighbour,accumulated pheromone level) for the 
    given saddr, daddr and species. */
void
CEASNetworkLayerUnit::getNeighboursCumulativePheroLevel(
  int32_t species,
  nsaddr_t src,
  nsaddr_t dst,
  vector<pair<NeighbourEntry*,double> >& neighbours)
{
    neighbourMap_t* neighbourhood = neighbourInformationBase_->getBase();
    double pherolevel_sum = 0;
    neighbourMap_t::iterator i = neighbourhood->begin();
    for (;  i != neighbourhood->end(); i++)
      { // for each neighbour entry
        NeighbourEntry* neighbour = dynamic_cast<NeighbourEntry*>(i->second);
        // if the neighbour is not reachable skip it
        if (!neighbour->isReachable()) continue;
        if (CEASPheroTableEntry* entry = pherotable()->findEntry(neighbour,
                                                                 species,
                                                                 src,
                                                                 dst))
          // there is an entry in the pherotable for this neighbour
          // add pheromone level towards neigbour
          pherolevel_sum += entry->getPheroLevel();
        // Store neigbour and accumulated pherolevel towards neigbour
        neighbours.push_back(
          pair<NeighbourEntry*,double>(neighbour,pherolevel_sum));
      }
}

/*! Build the list of (reachable neighbour,accumulated pheromone level) for the 
    given saddr, daddr, species and neighbour address. 
    Accumulated pheromone level is 0 for all neighbours when pheromone levels
    are ignored */
void
CEASNetworkLayerUnit::getNeighboursCumulativePheroLevel(
  int32_t species,
  nsaddr_t src,
  nsaddr_t dst,
  nsaddr_t neighbour_address,
  vector<pair<NeighbourEntry*,double> >& neighbours,
  bool ignorePherolevels)
{
    NeighbourKey key(neighbour_address,ANY_IFACE.value(),ANY_CHANNEL);
    list<NeighbourEntry*> entries;
    neighbourInformationBase_->lookupEntry(key,entries);
    double pherolevel_sum = 0;
    list<NeighbourEntry*>::iterator i = entries.begin();
    for (;  i != entries.end(); i++)
      { // for each neighbour entry
        NeighbourEntry* neighbour = dynamic_cast<NeighbourEntry*>(*i);
        // if the neighbour is not reachable, skip it
        if (!neighbour->isReachable()) continue;
        if (!ignorePherolevels)
          {
            if (CEASPheroTableEntry* entry = pherotable()->findEntry(neighbour,
                                                                     species,
                                                                     src,
                                                                     dst))
              // there is an entry in the pherotable for this neighbour
              // add pheromone level towards neigbour
              pherolevel_sum += entry->getPheroLevel();
          }
        // store neigbour and accumulated pherolevel towards neigbour
        neighbours.push_back(
          pair<NeighbourEntry*,double>(neighbour,pherolevel_sum));
      }
}

/*! \returns the reachable neighbours of given address on the given channel */
void
CEASNetworkLayerUnit::getNeighbours(
  nsaddr_t address,
  int channelIndex,
  vector<NeighbourEntry*>& neighbours)
{
    NeighbourKey key(address,ANY_IFACE.value(),channelIndex);
    list<NeighbourEntry*> entries;
    neighbourInformationBase_->lookupEntry(key,entries);
    if (!entries.empty())
      {
        list<NeighbourEntry*>::iterator i = entries.begin();
        for (;  i != entries.end(); i++)
          { // for each neighbour entry
            NeighbourEntry* neighbour = dynamic_cast<NeighbourEntry*>(*i);
            // if the neighbour is not reachable, skip it
            if (!neighbour->isReachable()) continue;
            // Store neigbour
            neighbours.push_back(neighbour);
          }
      }
}

/*! Find the best reachable node for the given saddr, daddr and flowid
    (specified in p)*/
NeighbourEntry*
CEASNetworkLayerUnit::getBestNeighbour(int32_t species,
                                       nsaddr_t src,
                                       nsaddr_t dst)
{
    NeighbourEntry* best_neighbour = 0;
    neighbourMap_t* neighbourhood = neighbourInformationBase_->getBase();
    double best_pherolevel = 0;
    neighbourMap_t::iterator i = neighbourhood->begin();
    for (;  i != neighbourhood->end(); i++)
      { // for each neighbour entry
        NeighbourEntry* neighbour = dynamic_cast<NeighbourEntry*>(i->second);
        // if the neighbour is not reachable, skip it
        if (!neighbour->isReachable()) continue;
        CEASPheroTableEntry* entry = pherotable()->findEntry(neighbour,
                                                             species,
                                                             src,
                                                             dst);
        if (entry && entry->getPheroLevel() > best_pherolevel )
          {
            // found best neighbour
            best_neighbour = neighbour;
            best_pherolevel = entry->getPheroLevel();
          }
      }
    return best_neighbour;
}

/*! Update pheromone levels for all entries in the pherotable for the given
    src,dst and species.*/
void
CEASNetworkLayerUnit::updatePhero(int32_t species,
                                  nsaddr_t src,
                                  nsaddr_t dst,
                                  NeighbourKey& s,
                                  double gamma,
                                  double L,
                                  double beta)
{
    neighbourMap_t* neighbourhood = neighbourInformationBase_->getBase();
    neighbourMap_t::iterator i = neighbourhood->begin();
    for (;  i != neighbourhood->end(); i++)
      { // for each neighbour entry
        NeighbourKey k = static_cast<NeighbourKey>(i->first);
        NeighbourEntry* neighbour = static_cast<NeighbourEntry*>(i->second);
        pherotable()->updateEntry(neighbour,species,src,dst,s==k,gamma,L,beta);
      }
}
