/*
 * Copyright (c) 2008 Q2S NTNU, Trondheim, Norway
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Port of the Swarm module developed by O. Wittner for ns-2.30.
 * Author: Laurent Paquereau <laurent.paquereau@q2s.ntnu.no>
 *
 * $Header$
 */

#ifndef _ceas_network_layer_unit_h
#define _ceas_network_layer_unit_h

#include <vector>
#include "network-layer-unit.h"
#include "network-interface.h"
#include "node.h"
#include "rng.h"
#include "ceas-neighbour-information-base.h"
#include "ceas-packet.h"
#include "ceas-pherotable.h"

#define PHERO_VERY_SMALL 10e-8

class CEASNetworkLayerUnit : public NetworkLayerUnit {
  public:
    CEASNetworkLayerUnit(Node* node);
    ~CEASNetworkLayerUnit();
    static inline const char* getTclConfig(const char* param)
    {
        Tcl& tcl = Tcl::instance();
        tcl.evalf("NetworkLayerUnit/CEAS set config_(%s)",param);
        return (tcl.result());
    }
    virtual inline packet_t packetType() const {return PT_CEAS;}
    virtual int command(int argc, const char*const* argv);
    //! ceas type handled
    virtual ceas_t ceasType() const = 0;
    inline nsaddr_t local_address() const {return node_->address();}
    virtual CEASPheroTable* pherotable() 
      {return static_cast<CEASPheroTable*>(rib_);}
    inline CEASNeighbourInformationBase* getNeighbourInformationBase() const
      {return neighbourInformationBase_;}
    void sendDown(Packet* p,Handler* h,NeighbourEntry* neighbour);
    virtual NeighbourEntry*
      selectRandomNeighbour(vector<pair<NeighbourEntry*,double> >& neighbours);
    void getAllNeighboursPheroLevel(
      int32_t species,
      nsaddr_t src,
      nsaddr_t dst,
      vector<pair<NeighbourEntry*,double> >& neighbours);
    void getNeighboursCumulativePheroLevel(
      int32_t species,
      nsaddr_t src,
      nsaddr_t dst,
      vector<pair<NeighbourEntry*,double> >& neighbours);
    template<class T> void getNeighboursCumulativePheroLevel(
      int32_t species,
      nsaddr_t src,
      nsaddr_t dst,
      vector<pair<NeighbourEntry*,double> >& neighbours,
      bool ignorePherolevels,
      list<pair<nsaddr_t,T> >& blacklist);
    void getNeighboursCumulativePheroLevel(
      int32_t species,
      nsaddr_t src,
      nsaddr_t dst,
      nsaddr_t neighbour_address,
      vector<pair<NeighbourEntry*,double> >& neighbours,
      bool ignorePherolevels);
    void getAllNeighbours(vector<pair<NeighbourEntry*,double> >& neighbours);
    void getNeighbours(nsaddr_t address,
                       int channelIndex,
                       vector<NeighbourEntry*>& neighbours);
    NeighbourEntry* getBestNeighbour(int32_t species,
                                     nsaddr_t src,
                                     nsaddr_t dst);
    void updatePhero(int32_t species,
                     nsaddr_t src,
                     nsaddr_t dst,
                     NeighbourKey& s,
                     double gamma,
                     double L,
                     double beta);
    template<class T> int count(nsaddr_t,list<pair<nsaddr_t,T> >&);
    template<class T> bool find(nsaddr_t,list<pair<nsaddr_t,T> >&);
  protected:
    virtual void addInterface(NetworkInterface2*);
    virtual void processPacket(Packet* p,Handler* h);
    CEASNeighbourInformationBase* neighbourInformationBase_;
    RNG* randomgen_; //!< random generator
};

/*! Build the list of (reachable neighbour,accumulated pheromone level) for the 
    given saddr, daddr and species. 
    Exclude nodes in blacklist.
    Accumulated pheromone level is 0 for all neighbours when pheromone levels
    are ignored */
template<class T> void
CEASNetworkLayerUnit::getNeighboursCumulativePheroLevel(
  int32_t species,
  nsaddr_t src,
  nsaddr_t dst,
  vector<pair<NeighbourEntry*,double> >& neighbours,
  bool ignorePherolevels,
  list<pair<nsaddr_t,T> >& blacklist)
{
    neighbourMap_t* neighbourhood = neighbourInformationBase_->getBase();
    double pherolevel_sum = 0;
    neighbourMap_t::iterator i = neighbourhood->begin();
    for (;  i != neighbourhood->end(); i++)
      { // for each neighbour entry
        NeighbourEntry* neighbour = dynamic_cast<NeighbourEntry*>(i->second);
        // if the neighbour is not reachable or in the blacklist skip it
        if (!neighbour->isReachable()||find(neighbour->getAddress(),blacklist))
          continue;
        if (!ignorePherolevels)
          {
            if (CEASPheroTableEntry* entry = pherotable()->findEntry(neighbour,
                                                                     species,
                                                                     src,
                                                                    dst))
              // there is an entry in the pherotable for this neighbour
              // add pheromone level towards neigbour
              pherolevel_sum += entry->getPheroLevel();
          }
        // store neigbour and accumulated pherolevel towards neigbour
        neighbours.push_back(
          pair<NeighbourEntry*,double>(neighbour,pherolevel_sum));
      }
}

/*! \returns the number of times addr appears in the list */
template<class T> int
CEASNetworkLayerUnit::count(nsaddr_t addr,list<pair<nsaddr_t,T> >& addrList)
{
    int i = 0;
    typename list<pair<nsaddr_t,T> >::iterator j = addrList.begin();
    for (;j != addrList.end(); j++)
      {
        if (j->first == addr) i++;
      }
    return i;
}

/*! \returns true if addr appears in the list, false otherwise */
template<class T> bool
CEASNetworkLayerUnit::find(nsaddr_t addr,list<pair<nsaddr_t,T> >& addrList)
{
    if (addrList.empty()) return false;
    typename list<pair<nsaddr_t,T> >::iterator j = addrList.begin();
    for (;j != addrList.end(); j++)
      {
        if (j->first == addr) return true;
      }
    return false;
}

#endif
