/*
 * Copyright (c) 2008 Q2S NTNU, Trondheim, Norway
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Port of the Swarm module developed by O. Wittner for ns-2.30.
 * Author: Laurent Paquereau <laurent.paquereau@q2s.ntnu.no>
 *
 * $Header$
 */
 
#ifndef _ceas_packet_generator_h
#define _ceas_packet_generator_h

#include <ip.h>
#include <routing-packet-generator.h>
#include "ceas-network-layer-unit.h"
#include "ceas-packet.h"

// Forward declaration
class CEASNetworkLayerUnit;
class Agent;
template<class _CEASData = CEASData<>* > class CEASPacketGenerator;

//! CEAS packet generator
template<class _CEASData>
class CEASPacketGenerator<_CEASData*> : public RoutingPacketGenerator {
  public:
    CEASPacketGenerator(CEASNetworkLayerUnit* networkLayerUnit);
    ~CEASPacketGenerator();
    virtual int command(int argc, const char*const* argv);
    virtual void expire(Event *); // Timeout handler
  protected:
    CEASAgentBehaviour_t behaviour_; //!< behaviour of CEAS packets
    int32_t species_; //!< species
    double beta_; //!< "evaporation" level for pheromones
    nsaddr_t destinationAddress_; //!< destination of the CEAS packets
    double rate_; //!< packet generation rate
    char* genType_; //!< generator type
    ceas_t ceasType_; //!< CEAS packet type
    RNG* randomgen_; //!< random generator
    virtual Packet* allocPacket() const;
};

//! Constructor
template<class _CEASData>
CEASPacketGenerator<_CEASData*>::CEASPacketGenerator(
  CEASNetworkLayerUnit* networkLayerUnit)
  : RoutingPacketGenerator(networkLayerUnit),
    behaviour_(CEAS_BEHAVIOUR_NORMAL),
    species_(-1),
    beta_(1),
    destinationAddress_(UNKN_ADDRESS),
    rate_(0.0),
    genType_(0),
    randomgen_(0)
{
    ceasType_ = networkLayerUnit->ceasType();
}

//! Destructor
template<class _CEASData>
CEASPacketGenerator<_CEASData*>::~CEASPacketGenerator()
{
    if (genType_) delete genType_;
}

//! Process commands fired at TCL-level
template<class _CEASData> int
CEASPacketGenerator<_CEASData*>::command(int argc, const char*const* argv)
{
    Tcl& tcl = Tcl::instance();
    if (argc == 3)
      {
        if (strcmp(argv[1], "set-species") == 0)
          { // Set species
            if(sscanf(argv[2], "%d", &species_) != 1)
              {
                tcl.resultf("CEASPacketGenerator::%s(set-species): "\
                            "Invalid pheromone id %s\n",__FUNCTION__,argv[2]);
                return (TCL_ERROR);
              }
            return (TCL_OK);
          }
        else if (strcmp(argv[1], "set-behaviour") == 0)
          { // Set behaviour
            if (strcmp(argv[2], "normal") == 0)
              behaviour_ = CEAS_BEHAVIOUR_NORMAL;
            else if (strcmp(argv[2], "explorer") == 0)
              behaviour_ = CEAS_BEHAVIOUR_EXPLORER;
            else
              {
                tcl.resultf("CEASPacketGenerator::%s(set-behaviour): "\
                            "Invalid behaviour type %s\n",__FUNCTION__,argv[2]);
                return (TCL_ERROR);
              }
            return (TCL_OK);
          }
        else if (strcmp(argv[1], "set-beta") == 0)
          { // Set "evaporation" level
            if (sscanf(argv[2], "%lf", &beta_) != 1)
              {
                tcl.resultf("CEASPacketGenerator::%s(set-beta): "\
                            "Invalid beta %s\n",__FUNCTION__,argv[2]);
                return (TCL_ERROR);
              }
            return (TCL_OK);
          }
        else if (strcmp(argv[1], "set-type") == 0)
          {// Set generator type
            if (genType_) delete genType_;
            genType_ = new char[strlen(argv[2])+1];
            strcpy(genType_, argv[2]);
            return (TCL_OK);
          }
        else if (strcmp(argv[1], "set-rate") == 0)
          { // Set packet generation rate.
            rate_ = atof(argv[2]);
            return (TCL_OK);
          }
        else if (strcmp(argv[1], "set-dst-address") == 0)
          { // Set destination address
            if (sscanf(argv[2], "%d", &destinationAddress_) != 1 ||
                destinationAddress_ == sourceAddress_ )
              {
                tcl.resultf("CEASPacketGenerator::%s(set-dst-address: "\
                            "Invalid dst address %s\n",__FUNCTION__,argv[2]);
                return (TCL_ERROR);
              }
            return (TCL_OK);
          }
        else if (strcmp(argv[1],"use-rng") == 0)
          { // set the random generator
            randomgen_ = dynamic_cast<RNG*>(TclObject::lookup(argv[2]));
            if (!randomgen_)
              {
                tcl.add_errorf("CEASPacketGenerator::%s (use-rng): "\
                               "Lookup failed (%s)",__FUNCTION__,argv[2]);
                return TCL_ERROR;
              }
            return TCL_OK;
          }        
      }
    return (RoutingPacketGenerator::command(argc, argv));
}

/*! Timeout handler. Create and send a new CEAS packet */
template<class _CEASData> void
CEASPacketGenerator<_CEASData*>::expire(Event *)
{
    if (!running_) return;

    // create packet data and packet
    _CEASData* data = new _CEASData(behaviour_,species_,beta_);
    packetSize_ = data->size() + IP_HDR_LEN + UDP_HDR_LEN;
    Packet* p = allocPacket();
    p->setdata(data);

    // send packet
    send(p,NULL);

    // schedule next packet
    if (strcmp(genType_, "CBR") == 0)
      { // contant bit rate implies fixed intervals
        if (rate_ > 0.0)
          resched(1.0/rate_);
      }
    else if (strcmp(genType_, "NED") == 0)
      { // negative exponential intevals
        if (rate_ > 0.0)
          resched(randomgen_->exponential()/rate_);
      }
    else
      {
        fprintf(stderr,
                "CEASPacketGenerator::%s: Unknown generator type (%s)\n",
                __FUNCTION__,genType_);
        abort();
      }
}

/*! Allocate a packet and fill in all the generic fields */
template<class _CEASData> Packet*
CEASPacketGenerator<_CEASData*>::allocPacket() const
{
    Packet* p = Packet::alloc();
    // common header
    hdr_cmn* ch = hdr_cmn::access(p);
    ch->uid() = Agent::uidcnt();
    ch->ptype() = packetType_;
    ch->size() = packetSize_;
    // IP header
    hdr_ip* ih = hdr_ip::access(p);
    ih->saddr() = sourceAddress_;
    if (destinationAddress_ != UNKN_ADDRESS) ih->daddr() = destinationAddress_;
    ih->ttl() = defttl_;
    ih->flowid() = species_;
    // CEAS header
    hdr_ceas* ceash = hdr_ceas::access(p);
    ceash->type() = ceasType_;
    return p;
}

#endif
