/*
 * Copyright (c) 2008 Q2S NTNU, Trondheim, Norway
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Port of the Swarm module developed by O. Wittner for ns-2.30.
 * Author: Laurent Paquereau <laurent.paquereau@q2s.ntnu.no>
 *
 * $Header$
 */
 
#ifndef _ceas_packet_h
#define _ceas_packet_h

#include <list>
#include "ns-process.h"
#include "packet.h"
#include "neighbour-information-base.h"

// CEAS agent state
enum CEASAgentState_t {
  CEAS_STATE_IDLE,
  CEAS_STATE_FORWARDSEARCH,
  CEAS_STATE_BACKTRACKING
};

// CEAS agent behaviour
enum CEASAgentBehaviour_t {
  CEAS_BEHAVIOUR_NORMAL,
  CEAS_BEHAVIOUR_EXPLORER
};

typedef unsigned int ceas_t;

static ceas_t CEAS_NONE = 0;

class CEAS_info {
  public:
    CEAS_info()
    {
        initName();
    }
    const char* name(ceas_t type) const
    { 
        if ( type <= CEAS_info::nCEAS_ ) return name_[type];
        return 0;
    }
    static void initName()
    {
        if(nCEAS_ >= CEAS_NONE+1)
          return;
        const char **nameNew = new const char*[CEAS_NONE+1];
        if (nCEAS_>1)
          {
            for(unsigned int i = 0; i < nCEAS_-1; i++)
              {
                nameNew[i] = name_[i];
              }
          }
        if(nCEAS_) delete [] name_;
        name_ = nameNew;
        nCEAS_ = CEAS_NONE+1;
        name_[CEAS_NONE]="none";
    }
    static int add(const char *name);
    static ceas_t type(const char *name)
    {
        for(unsigned int i = 0; i < nCEAS_; i++)
          {
            if(strcmp(name, name_[i]) == 0)
              return i;
          }
        return CEAS_NONE;
    }
  private:
    static const char** name_;
    static unsigned int nCEAS_;
};

//! CEAS packet header
struct hdr_ceas {
  ceas_t type_; //!< CEAS packet type
  static int offset_; //!< offset for this header
  inline static int& offset() { return offset_; }
  inline static hdr_ceas* access(Packet* p)
    {return (hdr_ceas*) p->access(offset_);}
  ceas_t& type() { return (type_); }
};

extern CEAS_info ceas_info; /* map CEAS_* to string name */
extern packet_t PT_CEAS;
extern AppDataType CEAS_DATA;

class CEASPathItemData;

//! Packet data holding state of CEAS agent
template<class _PathItemData = CEASPathItemData> 
class CEASData : public AppData {
  public:
    CEASData(CEASAgentBehaviour_t behaviour,int32_t species,double beta)
      : AppData(CEAS_DATA)
    {
        state_ = CEAS_STATE_IDLE;
        behaviour_ = behaviour;
        species_ = species;
        beta_ = beta;
        path_ = new list<pair<nsaddr_t,_PathItemData> >();
        gamma_ = 0;
        L_ = 0;
    }
    CEASData(CEASData& d) : AppData(d)
    {
        // Make this a copy of "d";
        state_ = d.state_;
        species_ = d.species_;
        beta_ = d.beta_;
        path_ = new list<pair<nsaddr_t,_PathItemData> >(d.path());
        gamma_ = d.gamma_;
        L_ = d.L_;
    }
    virtual ~CEASData()
    {
        delete path_;
    }

    // Access methods
    inline CEASAgentState_t& state() {return state_; };
    inline CEASAgentBehaviour_t& behaviour() {return behaviour_;};
    inline int32_t& species() {return species_; };
    inline list<pair<nsaddr_t,_PathItemData> >& path() {return *path_;};
    inline double& gamma() {return gamma_;};
    inline double& beta() {return beta_;};
    inline double& L() {return L_;};
    virtual int size() const
    {
        return 4 /* ceas header + state + behaviour + species */
          + path_->size() * (4 /*nsaddr_t*/ + path_->front().second.size())  
          + 12 /* gamma, beta, L, assume float size */;
    }
    virtual AppData* copy() { return new CEASData(*this); }
    inline nsaddr_t& nest_addr() {return path_->front().first;}
    virtual void push(nsaddr_t,NeighbourEntry*);
    virtual void pop();
    inline bool isBackAtNest() 
    {
        if (state_ == CEAS_STATE_BACKTRACKING && path_->empty())
          return true;
        return false;
    }
    bool isAtDestination(nsaddr_t destination,nsaddr_t current_node);
  protected:
    // CEAS agent state info
    CEASAgentState_t state_; //!< State of agent
    CEASAgentBehaviour_t behaviour_; //!< Behaviour of the agent
    int32_t species_;             //!< CEAS agent species id
    /*! Path. A path item is a pair (node address,data).
        Forward ant: path followed so far.
        Backward ant: path yet to follow. */
    list<pair<nsaddr_t,_PathItemData> >* path_;
    double gamma_; //!< "Temperature" of elite ants
    double beta_; //!< "Evaporation factor"
    double L_; //!< Path cost
};

template<class _PathItemData> void
CEASData<_PathItemData>::pop()
{
    assert(!path_->empty());
    path().pop_back();
}

template<class _PathItemData> bool
CEASData<_PathItemData>::isAtDestination(nsaddr_t destination,
                                         nsaddr_t current_node)
{
    if (destination != current_node) return false;
    switch (state_) {
      case CEAS_STATE_FORWARDSEARCH:
        {
          if (!path_->empty()) return true;
          break;
        }
      case CEAS_STATE_BACKTRACKING:
        {
          NeighbourKey unkn_key(UNKN_ADDRESS,UNKN_IFACE.value(),UNKN_CHANNEL);
          if (path_->back().second.next() == unkn_key)
            return true;
          break;
        }
      default:
          break;
    }
    return false;
}

//! Simplest _PathItemData; only contains the neighbour (next-hop)
class CEASPathItemData {
  public:
    CEASPathItemData(NeighbourKey next):next_(next) {}
    inline NeighbourKey& next() {return next_;}
    virtual int size() const {return 8 /* assume next_ is 8 bytes*/;}
  protected:
    NeighbourKey next_; //!< next-hop key
};

#endif
