/*
 * Copyright (c) 2008 Q2S NTNU, Trondheim, Norway
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Port of the Swarm module developed by O. Wittner for ns-2.30.
 * Author: Laurent Paquereau <laurent.paquereau@q2s.ntnu.no>
 *
 * $Header$
 */
 
#include "ceas-pherotable.h"
#include <math.h>
#ifndef stand_alone
#include "ceas-network-layer-unit.h"

// OTcl linkage
static class CEASPheroTableClass : public TclClass {
public:
    CEASPheroTableClass()
      : TclClass("NetworkInformationBase/RIB/CEASPheroTable") {}
    TclObject* create(int argc, const char*const* argv)
    {
        if (argc != 5)
          { // 5 <=> 1 parameter for the the constructor
            fprintf(stderr,"CEASPheroTableClass::%s: Invalid number of "\
                    "parameters.\n",__FUNCTION__);
            abort();
          }
        Node* node = dynamic_cast<Node*>(TclObject::lookup(argv[4]));
        if (!node)
          { // lookup failed
            fprintf(stderr,
                    "CEASPheroTableClass::%s: Lookup failed (node)(%s)\n",
                    __FUNCTION__,argv[4]);
            abort();
          }
         CEASPheromone_t pheromoneType = static_cast<CEASPheromone_t>(atoi(
           CEASNetworkLayerUnit::getTclConfig("pheromone-approx")));
         switch (pheromoneType) {
           case CEAS_PHEROMONE_SECOND_ORDER_HYPERBOLIC_APPROX:
             return (new CEASPheroTableImpl<CEASPheroTableEntrySecondOrderHyperbolic*>(node));
           case CEAS_PHEROMONE_FIRST_ORDER_LINEAR_APPROX:
             return (new CEASPheroTableImpl<CEASPheroTableEntryFirstOrderLinear*>(node));
           case CEAS_PHEROMONE_SECOND_ORDER_LINEAR_APPROX:
             return (new CEASPheroTableImpl<CEASPheroTableEntrySecondOrderLinear*>(node));
           default:
             break;
        }
        fprintf(stderr,"CEASPheroTableClass::%s: Invalid pheromone type\n",
                __FUNCTION__);
        abort();
    }
} class_CEASPheroTable;

//! Constructor
CEASPheroTableEntry::CEASPheroTableEntry(char* dst,
                                         NeighbourEntry* neighbour,
                                         nsaddr_t src,
                                         int32_t species,
                                         int preference)
  : NetworkInformationBaseEntry(dst,
                                neighbour->getInterface(),
                                neighbour->getAddress(),
                                neighbour->getChannel(),
                                src,
                                species,
                                false,
                                UNKN_METRIC,
                                preference),
#else
CEASPheroTableEntry::CEASPheroTableEntry() :
#endif
    pherolevel_(0)
{
}

//! Update entry
void
CEASPheroTableEntrySecondOrderHyperbolic::update(bool I,
                                                 double gamma,
                                                 double L,
                                                 double beta)
{
    double L_on_gamma = L/gamma;
    double H = exp(-L_on_gamma );
    //update the pheromone level
    pherolevel_ = (I ? H : 0) + A_ + (B_*C_ == 0 || gamma > (2*C_/B_) ?
                 - B_/gamma + C_/pow(gamma,2) : -pow(B_,2)/(4*C_) );
    // check for underflow
    if (pherolevel_ < 0)
      { // reset all parameters
        pherolevel_ = 0;
        A_ = 0;
        B_ = 0;
        C_ = 0;
      }
    else
      { // update pheromone autoreg parameters
        A_ = beta * (A_ + (I ? H*(1+L_on_gamma*(1+L_on_gamma/2)) : 0));
        B_ = beta * (B_ + (I ? H*L*(1+L_on_gamma) : 0));
        C_ = beta * (C_ + (I ? H*pow(L,2)/2 : 0));
      }
}

//! Update entry
void
CEASPheroTableEntryFirstOrderLinear::update(bool I,
                                            double gamma,
                                            double L,
                                            double beta)
{
    double L_on_gamma = L/gamma;
    double H = exp(-L_on_gamma );
    t_++;
    //update the pheromone level
    double upper_limit = beta * (1-pow(beta,t_-1))/(1-beta);
    double prev = A_ + B_* gamma;
    if (prev > upper_limit) { prev = upper_limit; A_ = upper_limit; B_ = 0; printf("Warning: above max\n");}
    else if (prev < 0) { prev = 0; A_ = 0; B_ = 0;printf("Warning: below 0\n");}
    pherolevel_ = (I ? H : 0) + prev;
    // check for underflow
    if (pherolevel_ < 0)
      { // reset all parameters
        pherolevel_ = 0;
        A_ = 0;
        B_ = 0;
      }
    else
      { // update pheromone autoreg parameters
        A_ = beta * (A_ + (I ? H*(1-L_on_gamma) : 0));
        B_ = beta * (B_ + (I ? H*L_on_gamma/gamma : 0));
      }
}

//! Update entry
void
CEASPheroTableEntrySecondOrderLinear::update(bool I,
                                             double gamma,
                                             double L,
                                             double beta)
{
    double L_on_gamma = L/gamma;
    double H = exp(-L_on_gamma );
    t_++;
    double upper_bound = beta * (1-pow(beta,t_-1))/(1-beta);
    //update the pheromone level
    pherolevel_ = A_ +
                  ((C_>0 && gamma<B_/C_) || (C_<0 && gamma>B_/C_) ?
                   -pow(B_,2)/(2*C_) : -B_*gamma+C_/2*pow(gamma,2));
    if (pherolevel_ < 0)
      { // reset all parameters
        pherolevel_ = 0;
        A_ = 0;
        B_ = 0;
        C_ = 0;
      }
    else if (pherolevel_ > upper_bound) 
        pherolevel_= upper_bound;
    if (I) pherolevel_ += H;
    // update pheromone autoreg parameters*/
    A_ = beta * (A_ + (I ? H*(1+L_on_gamma*(L_on_gamma/2-2)) : 0));
    B_ = beta * (B_ + (I ? H*L_on_gamma/gamma*(L_on_gamma-3) : 0));
    C_ = beta * (C_ + (I ? H*L_on_gamma/pow(gamma,2)*(L_on_gamma-2) : 0));
}
