/*
 * Copyright (c) 2008 Q2S NTNU, Trondheim, Norway
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Port of the Swarm module developed by O. Wittner for ns-2.30.
 * Author: Laurent Paquereau <laurent.paquereau@q2s.ntnu.no>
 *
 * $Header$
 */

#ifndef _ceas_pherotable_h
#define _ceas_pherotable_h

#include <map>
#ifndef stand_alone
#include <config.h>
#include <address.h>
#include <channel.h>
#include <neighbour-information-base.h>
#include <network-information-base.h>
#include <network-layer-unit.h>
#include <point-to-point-interface.h>
#endif

#define ANY_SPECIES -2 // wildcard

enum CEASPheromone_t {
  CEAS_PHEROMONE_SECOND_ORDER_HYPERBOLIC_APPROX = 0,
  CEAS_PHEROMONE_FIRST_ORDER_LINEAR_APPROX = 1,
  CEAS_PHEROMONE_SECOND_ORDER_LINEAR_APPROX = 2,
};

//! Pheromone table entry
class CEASPheroTableEntry 
#ifndef stand_alone
: public NetworkInformationBaseEntry 
#endif
{
  public:
#ifndef stand_alone
    CEASPheroTableEntry(char* dst,
                        NeighbourEntry* neighbour,
                        nsaddr_t src,
                        int32_t species,
                        int preference);
#else
    CEASPheroTableEntry();
#endif
    inline double getPheroLevel() const {return pherolevel_;};
    virtual void update(bool I,double gamma,double L,double beta)=0;
  protected:
    double pherolevel_; //!< pheromone level
};

//! Use second order hyperbolic approximation
class CEASPheroTableEntrySecondOrderHyperbolic : public CEASPheroTableEntry {
  public:
#ifndef stand_alone
    CEASPheroTableEntrySecondOrderHyperbolic(char* dst,
                                             NeighbourEntry* neighbour,
                                             nsaddr_t src,
                                             int32_t species,
                                             int preference)
      : CEASPheroTableEntry(dst,neighbour,src,species,preference)
#else
    CEASPheroTableEntrySecondOrderHyperbolic() : CEASPheroTableEntry()
#endif
      ,A_(0),B_(0),C_(0) {}
    virtual void update(bool I,double gamma,double L,double beta);
  protected:
    double A_; //!< auto-regression parameter
    double B_; //!< auto-regression parameter
    double C_; //!< auto-regression parameter
};

//! Use first order linear approximation
class CEASPheroTableEntryFirstOrderLinear : public CEASPheroTableEntry {
  public:
#ifndef stand_alone
    CEASPheroTableEntryFirstOrderLinear(char* dst,
                                        NeighbourEntry* neighbour,
                                        nsaddr_t src,
                                        int32_t species,
                                        int preference)
      : CEASPheroTableEntry(dst,neighbour,src,species,preference)
#else
    CEASPheroTableEntryFirstOrderLinear() : CEASPheroTableEntry()
#endif
      ,A_(0),B_(0),t_(0) {}
    virtual void update(bool I,double gamma,double L,double beta);
  protected:
    double A_; //!< auto-regression parameter
    double B_; //!< auto-regression parameter
    unsigned int t_;//!< number of iterations
};

//! Use second order linear approximation
class CEASPheroTableEntrySecondOrderLinear : public CEASPheroTableEntry {
  public:
#ifndef stand_alone
    CEASPheroTableEntrySecondOrderLinear(char* dst,
                                         NeighbourEntry* neighbour,
                                         nsaddr_t src,
                                         int32_t species,
                                         int preference)
      : CEASPheroTableEntry(dst,neighbour,src,species,preference)
#else
    CEASPheroTableEntrySecondOrderLinear() : CEASPheroTableEntry()
#endif
      ,A_(0),B_(0),C_(0),t_(0) {}
    virtual void update(bool I,double gamma,double L,double beta);
  protected:
    double A_; //!< auto-regression parameter
    double B_; //!< auto-regression parameter
    double C_; //!< auto-regression parameter
    unsigned int t_;//!< number of iterations
};


#ifndef stand_alone

class CEASPheroTable : public RIB {
  public:
    CEASPheroTable(Node* node) : RIB(node) {}
    virtual void updateEntry(NeighbourEntry* neighbour,
                             int32_t species,
                             nsaddr_t src,
                             nsaddr_t dst,
                             bool I,
                             double gamma,
                             double L,
                             double beta) = 0;
    virtual CEASPheroTableEntry* findEntry(NeighbourEntry* neighbour,
                                           int32_t species,
                                           nsaddr_t src,
                                           nsaddr_t dst) = 0;
};

template<class _CEASPheroTableEntry> class CEASPheroTableImpl;

//! Pheromone table for CEAS NetworkLayerUnit
template<class _CEASPheroTableEntry>
class CEASPheroTableImpl<_CEASPheroTableEntry*> : public CEASPheroTable {
  public:
    CEASPheroTableImpl(Node* node) : CEASPheroTable(node) {}
    virtual void updateEntry(NeighbourEntry* neighbour,
                             int32_t species,
                             nsaddr_t src,
                             nsaddr_t dst,
                             bool I,
                             double gamma,
                             double L,
                             double beta);
    virtual _CEASPheroTableEntry* findEntry(NeighbourEntry* neighbour,
                                            int32_t species,
                                            nsaddr_t src,
                                            nsaddr_t dst);
  protected:
    _CEASPheroTableEntry* addEntry(NeighbourEntry* neighbour,
                                   int32_t species,
                                   nsaddr_t src,
                                   nsaddr_t dst);
    _CEASPheroTableEntry* getEntry(NeighbourEntry* neighbour,
                                   int32_t species,
                                   nsaddr_t src,
                                   nsaddr_t dst);
  protected:
    virtual const char* getBaseName() {return "CEAS Pheromone Table";}
};

/*! Add entry in the pheromone table for (src,dst,species,iface) and return it
    Erase any existing entry for this tuple*/
template<class _CEASPheroTableEntry> _CEASPheroTableEntry*
CEASPheroTableImpl<_CEASPheroTableEntry*>::addEntry(NeighbourEntry* neighbour,
                                                    int32_t species,
                                                    nsaddr_t src,
                                                    nsaddr_t dst)
{
    char* dststr = addr2str(dst);
    _CEASPheroTableEntry* entry = new _CEASPheroTableEntry(
      dststr,
      neighbour,
      src,
      species,
      networkLayerUnit_->getPreference());
    RIB::addEntry(dststr,entry);
    delete[] dststr;
    return entry;
}

/*! \returns entry for neighbour, species, src and dst if any found, 0 otherwise
    Raise an error if more than one entry are found */
template<class _CEASPheroTableEntry> _CEASPheroTableEntry*
CEASPheroTableImpl<_CEASPheroTableEntry*>::findEntry(NeighbourEntry* neighbour,
                                                     int32_t species,
                                                     nsaddr_t src,
                                                     nsaddr_t dst)
{
    DestinationIBKey key(
      neighbour->getInterface()->getId(),
      neighbour->getAddress(),
      neighbour->getChannel() ? neighbour->getChannel()->index() : ANY_CHANNEL,
      src,
      species);
    list<NetworkInformationBaseEntry*> entries;
    char* dststr = addr2str(dst);  
    lookupEntry(dststr,key,false,entries);
    delete[] dststr;
    if(entries.empty())
      return 0;
    else if (entries.size()==1)
      return (dynamic_cast<_CEASPheroTableEntry*>(entries.front()));
    else
      {
        fprintf(stderr,
                "PheroTable::%s: Error - more than one entry found\n",
                __FUNCTION__);
        abort();
      }
}

/*! \returns entry for neighbour, species, src and dst. If no entry exist,
    add one. If more than one entry are found, raise an error */
template<class _CEASPheroTableEntry> _CEASPheroTableEntry*
CEASPheroTableImpl<_CEASPheroTableEntry*>::getEntry(NeighbourEntry* neighbour,
                                                    int32_t species,
                                                    nsaddr_t src,
                                                    nsaddr_t dst)
{
    _CEASPheroTableEntry* entry = findEntry(neighbour,species,src,dst);
    if (!entry) entry = addEntry(neighbour,species,src,dst);
    return entry;
}

/*! update the entry for the given neighbour, species, src, dst with I, gamma, L
    and beta*/
template<class _CEASPheroTableEntry> void
CEASPheroTableImpl<_CEASPheroTableEntry*>::updateEntry(
  NeighbourEntry* neighbour,
  int32_t species,
  nsaddr_t src,
  nsaddr_t dst,
  bool I,
  double gamma,
  double L,
  double beta)
{
    _CEASPheroTableEntry* entry = getEntry(neighbour,species,src,dst);
    entry->update(I,gamma,L,beta);
}

#endif

#endif
