/*
 * Copyright (c) 2008 Q2S NTNU, Trondheim, Norway
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Port of the Swarm module developed by O. Wittner for ns-2.30.
 * Author: Laurent Paquereau <laurent.paquereau@q2s.ntnu.no>
 *
 * $Header$
 */

#ifndef _ceas_routing_unit_h
#define _ceas_routing_unit_h

#include <limits>
#include "drop.h"
#include "neighbour-information-base.h"
#include "routing-unit.h"
#include "ceas-adib.h"
#include "ceas-network-layer-unit.h"
#include "ceas-packet-generator.h"

#define CEAS_ELITE_SELECTION_EPSILON  10e-12
//#define CEAS_ELITE_SELECTION_EPSILON 2*numeric_limits<double>::epsilon()

//! Elite selection strategies
enum CEASEliteSelection_t {
  // do not apply elite selection
  CEAS_ELITE_SELECTION_OFF = 0,
  // apply elite selection on non-explorer ants only
  CEAS_ELITE_SELECTION_ON_NON_EXPLORER = 1,
  // apply elite selection on all ants
  CEAS_ELITE_SELECTION_ON_ALL = 2
};

//! Routing loops policy
enum CEASRoutingLoops_t {
  // discard forward ant when routing loop
  CEAS_RT_LOOPS_PREVENT = 0,
  // allow routing loops
  CEAS_RT_LOOPS_ALLOW = 1,
  // remove loops at destination
  CEAS_RT_LOOPS_REMOVE_AT_DESTINATION = 2
};

// Forward declarations
template<class __Key,class __PathItemData> class CEAS_ADIB;
template<class _Key,class _PathItemData> class CEASLogger;

//! CEAS Routing Unit
template<class _Key = int32_t,class _PathItemData = CEASPathItemData>
class CEASRoutingUnit : public RoutingUnit {
  public:
    CEASRoutingUnit(CEASNetworkLayerUnit* networkLayerUnit);
    ~CEASRoutingUnit();
    inline CEASNetworkLayerUnit* networkLayerUnit()
      {return static_cast<CEASNetworkLayerUnit*>(networkLayerUnit_);}
    virtual void recv(Packet* p,Handler *h);
    virtual void run() = 0;
    inline Packet* p() {return p_;}
    _Key& key() {return key_;}
    CEASData<_PathItemData>* ant() const {return ant_;}
    virtual int32_t species() {return ant_->species();} 
    inline nsaddr_t local_address(){return networkLayerUnit()->local_address();}
    inline CEASRoutingUnitIB<_Key,CEAS_ADIBEntry<_Key,CEASTemperature*,_PathItemData>*,_PathItemData>&
      adib() {return *CEASRoutingUnit<_Key,_PathItemData>::adib_;}
    virtual CEASLogger<_Key,_PathItemData>* logger()
      {return static_cast<CEASLogger<_Key,_PathItemData>*>(logger_);}
    inline bool eliteSelection() const {return (bool)eliteSelection_;}
  protected:
    Packet* p_;  //!< Latest packet
    Handler* h_; //!< Latest handler
    CEASData<_PathItemData>* ant_; //!< Latest CEAS agent
    _Key key_; //! Latest key
    CEAS_ADIB<_Key,_PathItemData>* adib_; //! At Destination IB
    CEASEliteSelection_t eliteSelection_; //!< elite selection policy
    CEASRoutingLoops_t routingLoops_; //!< routing loop handling policy
    bool rhoReduction_; //!< true if rho reduction is enabled
    inline double updateTotalTemperature();
    bool doEliteSelection(double& gamma);
    void doRhoReduction();
    virtual void sendTo(NeighbourEntry*);
    void idle2Forward();
    void forward2Backtracking();
    template<class T> static void removeLoops(list<pair<nsaddr_t,T> >* path);
};

//! Constructor
template<class _Key,class _PathItemData>
CEASRoutingUnit<_Key,_PathItemData>::CEASRoutingUnit(CEASNetworkLayerUnit* networkLayerUnit)
  : RoutingUnit(networkLayerUnit),p_(0),h_(0),ant_(0)
{
    // fetch the configuration parameters from tcl
    eliteSelection_ = static_cast<CEASEliteSelection_t>(atoi(
      CEASNetworkLayerUnit::getTclConfig("elite-selection")));
    routingLoops_ = static_cast<CEASRoutingLoops_t>(atoi(
      CEASNetworkLayerUnit::getTclConfig("routing-loops")));
    rhoReduction_ =atoi(CEASNetworkLayerUnit::getTclConfig("rho-reduction"))==1;  
    // create information base
    adib_ = new CEAS_ADIB<_Key,_PathItemData>(this);
}

//! Destructor
template<class _Key,class _PathItemData>
CEASRoutingUnit<_Key,_PathItemData>::~CEASRoutingUnit()
{
    if (adib_) delete adib_;
}

template<class _Key,class _PathItemData> void
CEASRoutingUnit<_Key,_PathItemData>::recv(Packet* p,Handler *h)
{
    p_ = p;
    hdr_cmn* ch = hdr_cmn::access(p_);
    hdr_ip*  ih = hdr_ip::access(p_);
    if (ch->direction() == hdr_cmn::UP) 
      // decrement ttl
      ih->ttl()--;
    h_ = h;
    ant_ = static_cast<CEASData<_PathItemData>*>(p_->userdata());
    run();
}

template<class _Key,class _PathItemData> void
CEASRoutingUnit<_Key,_PathItemData>::sendTo(NeighbourEntry* next)
{
    hdr_cmn* ch = hdr_cmn::access(p_);
    hdr_ip*  ih = hdr_ip::access(p_);
    if (ch->direction() == hdr_cmn::DOWN && ih->ttl()==0)
      { // ttl is 0, drop packet
        networkLayerUnit()->drop(p_,DROP_RTR_TTL);
        return;
      }
    // update packet size
    ch->size() = ant()->size()+IP_HDR_LEN + UDP_HDR_LEN;
    // send
    networkLayerUnit()->sendDown(p_,h_,next);
}

/*! Update total temperature for the path of id key_
   \returns the updated total gamma */
template<class _Key,class _PathItemData> double
CEASRoutingUnit<_Key,_PathItemData>::updateTotalTemperature()
{
    double gamma = (*adib_)[key_]->updateTotalTemperature(ant()->L(),
                                                          ant()->beta());
    if (logger()) 
      logger()->sample(this,CEAS_LOG_TOTAL_GAMMA|CEAS_LOG_TOTAL_TEMPERATURE_VARS);
    return gamma;
}

/*! If elite selection is enabled, update the elite temperature of the path
    of id key_ if the path is in the elite set.
    \param gamma total temperature of the path of id key_; it is changed to
                 the elite temperature is the path is in the elite set, to -1
                 otherwise.
    \returns true if the path of id key_ should be updated */
template<class _Key,class _PathItemData> bool
CEASRoutingUnit<_Key,_PathItemData>::doEliteSelection(double& gamma)
{
    bool backtrack = false;
    switch (eliteSelection_) {
      case CEAS_ELITE_SELECTION_OFF:
        return true;
      case CEAS_ELITE_SELECTION_ON_NON_EXPLORER:
        if (ant()->behaviour() == CEAS_BEHAVIOUR_EXPLORER)
          backtrack = true;
      case CEAS_ELITE_SELECTION_ON_ALL:
        if (!(ant()->L() > (*adib_)[key_]->eliteSelectionLevel()) ||
            fabs(ant()->L() - (*adib_)[key_]->eliteSelectionLevel() <= 
            CEAS_ELITE_SELECTION_EPSILON))
          { // the path is in the "elite set"; update elite temperature
            gamma = (*adib_)[key_]->updateEliteTemperature(ant()->L(),
                                                           ant()->beta());
            backtrack = true;
            if (logger()) logger()->sample(this,CEAS_LOG_ELITE_GAMMA);
          }
        else if (backtrack)
          gamma = (*adib_)[key_]->eliteTemperature()->gamma();
        else
          gamma = -1;
      default:
        break;
    }
    if (logger()) logger()->sample(this,CEAS_LOG_ELITE_SELECTION_LEVEL);
    return backtrack;
}

/*! If rho reduction is enabled, update rho for the path of id key_ */
template<class _Key,class _PathItemData> void
CEASRoutingUnit<_Key,_PathItemData>::doRhoReduction()
{
    if (!rhoReduction_) return;
    (*adib_)[key_]->updateRho(ant()->L());
}

/*! Change the ant from idle to forward state and make the necessary header
    updates */
template<class _Key,class _PathItemData> void
CEASRoutingUnit<_Key,_PathItemData>::idle2Forward()
{
    assert(ant()->state()==CEAS_STATE_IDLE);
    ant()->state() = CEAS_STATE_FORWARDSEARCH;
    if (logger())  logger()->sample(this,CEAS_LOG_FORWARDCOUNT);
}

/*! Change the ant from forward to backtracking state and make the necessary
    header updates */
template<class _Key,class _PathItemData> void
CEASRoutingUnit<_Key,_PathItemData>::forward2Backtracking()
{
    assert(ant()->state()==CEAS_STATE_FORWARDSEARCH);
    ant()->state() = CEAS_STATE_BACKTRACKING;
    // update IP header
    hdr_ip*  ih = hdr_ip::access(p_);
    ih->saddr() = local_address();
    ih->daddr() = ant()->nest_addr();
    ih->ttl() = ant()->path().size()-1;
    if (logger()) logger()->sample(this,CEAS_LOG_BACKTRACKCOUNT);
}


/*! Make the given path loop-free.
   NB: when T is a pointer, objects are removed from the list but not deleted*/
template<class _Key,class _PathItemData> template<class T> void
CEASRoutingUnit<_Key,_PathItemData>::removeLoops(list<pair<nsaddr_t,T> >* path)
{
    typename list<pair<nsaddr_t,T> >::iterator i = path->begin();
    typename list<pair<nsaddr_t,T> >::iterator j = path->begin();
    while (i != path->end())
      {
        while (j != path->end())
          {
            j++;
            if (j == path->end()) break;
            if (i->first == j->first)
              {
                path->erase(i,j);
                i = path->begin();
                j = path->begin();
              }
          }
        i++;
        j = i;
      }
}

#endif
