# Copyright (c) 2008 Q2S NTNU, Trondheim, Norway
# All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation;
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# Port of the Swarm module developed by O. Wittner for ns-2.30.
# Author: Laurent Paquereau <laurent.paquereau@q2s.ntnu.no>
#
# $Header$

#
# Dynamic library helpers
# (copied from nsmiracle-1.2.X/samples/dynlibutils.tcl, Copyright SIGNET)
#

set dynlibutils_system_type [exec uname -s]

proc dynlibname {name} {
   global dynlibutils_system_type
    # Libraries have different names in some operating systems
    if {[string match "CYGWIN*" "$dynlibutils_system_type"] == 1} {
        return cyg${name}-0.dll
    } else {
        return lib${name}.so
    } 

}

proc dynlibload {name {path ""}} {
    if {$path == ""} {
        load "[dynlibname ${name}]"
    } else    {
        load "${path}/[dynlibname ${name}]"
    }
}


#
# Debug stuff
#

proc DEBUG text {
    # Output debug text
    global param
    if { $param(DEBUG_ON) } {
        puts "DEBUG $text"
    }
}

proc DEBUGA { text arr } {
    # Output debug text and array
    global param
    if { $param(DEBUG_ON) } {
        puts "DEBUGA $text"
        array set x $arr
        parray x
    }
}


global param CEASPath

# Load CEAS common library
dynlibload ceascommon

# Source topology
if { [ info exist param(Topology) ] } {
    source $CEASPath/topologies/$param(Topology)
} else {
    puts "CEAS_scenario: Error - No topology specified"
    exit 1
}   


# 
# Base class for CEAS scenarios in NS
#

Class CEAS_scenario

# Constructor
# - param is expected to be in "array get" format
CEAS_scenario instproc init { param } {
    $self instvar ns_ param_ 
    set ns_ [Simulator instance]
    array set param_ $param
    puts "Init simulation..."
    $self init_simulation  ;# Misc. init related to sim.
    puts "Init topology..."
    $self init_topology    ;# Prepare base topology
    puts "Init components..."
    $self init_components  ;# Init. misc. components
    puts "Schedule events..."
    $self schedule_events  ;# Schedule simulation events
    puts "Init completed."
}

# Perform misc initialisation related to monitoring
# the simulation
CEAS_scenario instproc init_simulation {} {
    $self load_ext_param   ;# Load external parameters if required
    $self dump_param       ;# Dump parameters to file
    $self get_simid        ;# Get id of simulation 
    $self setup_trace      ;# Setup tracing
    $self config_global    ;# Set some global configuration
}

CEAS_scenario instproc config_global {} {
    $self instvar ns_
    # Do not use any rtproto (ns-2 default is Static)
    $ns_ rtproto None
    # Set network layer manager
    $ns_ node-config -network-layer-manager on
    # Set type of network layer
    global network_layer_type
    if { [llength [split [NetworkLayerUnit/CEAS info subclass]]] == 1 } {
         set network_layer_type [string replace [NetworkLayerUnit/CEAS info subclass] 0 16]
    }
}

# Prepare topology based om values set in param
CEAS_scenario instproc init_topology {} {
    $self instvar ns_ param_ numnodes_ node_ ceas_network_layer_unit_ nam_f_
    $self get_colors       ;# Define list of colors  
    # Create network topology
    # Create nodes and links
    # global variables produced: node NumNodes network_layer_unit
    create_net;
    global NumNodes node network_layer_unit
    set numnodes_ $NumNodes; # Store of no of nodes in topology
    set param_(numnodes) $numnodes_
    array set node_ [array get node]
    array set ceas_network_layer_unit_ [array get network_layer_unit]
}

# Prepare misc. component required in simulation
CEAS_scenario instproc init_components {} {
    puts "CEAS_scenario: Warning - No components initialized"
}

# Schedule events
CEAS_scenario instproc schedule_events {} {
    $self instvar ns_ param_ next_stop_
    $ns_ at 0.0 "$self show_progress"
    if { ! [info exist param_(Finish)] } {
        puts "CEAS_scenario: Warning - No finish event. Set to time 0.0."
        set param_(Finish) 0.0
    }
    if { [info exist param_(StopToContinue)] && 
        $param_(StopToContinue)} {
        # Enable continue after finish
        set next_stop_ $param_(Finish)
        $ns_ at $next_stop_ "$self stop_and_continue"
    } else {
        $ns_ at $param_(Finish) "$self finish"
    }
}    

# Load parameters from file and override/merge with current parameter
CEAS_scenario instproc load_ext_param {} {
    $self instvar ns_ param_
    if { [ info exist param_(LoadParam) ] } {
        set paramfile [ open $param_(LoadParam) r]
        while { ! [eof $paramfile]} {
            set p [ gets $paramfile ]
            scan "$p" "param(%s = %s" p1 p2
            set p1 [string trimright $p1 ")"]
            # Skip "LoadParam" parameter
            if { $p1 != "LoadParam" } {
                set param_($p1) $p2
            }
        }
    }
}

# Extract id of simulation from name of current directory
CEAS_scenario instproc get_simid {} {
    global argc argv
    $self instvar SimID_
    if { [ info exist SimID_ ] } {
        return
    }
    if { $argc == 1 } {
      set SimID_ [lindex $argv 0]
      # Use simply current dir as simulation id
      puts "Simulation ([pid]) results output to '[lindex [split [pwd] /] end]'"
    } else {
      scan [lindex [split [pwd] /] end] "simrun%d" SimID_
      if { ! [ info exist SimID_ ] } {
          # Use simply current dir as simulation id
          set SimID_ [lindex [split [pwd] /] end]
          puts "Simulation ([pid]) results output to '$SimID_/'"
      } else {
          puts "Simulation ([pid]) results output to 'simrun$SimID_/'"
      }
   }
   if ![string is integer -strict $SimID_] {
     puts stderr "Invalid simulation id ($SimID_). Aborting..."
     exit 1
   }
}

# Log parameters to file
CEAS_scenario instproc dump_param {} {
    $self instvar param_ 
    set param_file [open "parameters.log" w]
    set elements [array names param_]
    foreach i $elements {
        puts $param_file "param($i) = $param_($i)"
    }
    # Flush $param_file
    close $param_file
}

# Setup simulator output
CEAS_scenario instproc setup_trace {} {
    $self instvar ns_ param_ trace_f_ nam_f_
    # Open trace files
    if { [info exist param_(TRACE_ON)] && $param_(TRACE_ON) } {
        if { ! [info exist param_(TraceFile)] } {
            puts "CEAS_scenario: Warning - No trace file specified. Disabling trace."
            set param_(TRACE_ON) 0
        } else {
           set trace_f_ [open $param_(TraceFile) w]
           $ns_ trace-all $trace_f_
        }
    }
    if { [info exist param_(NAMTRACE_ON) ] && $param_(NAMTRACE_ON) } {
        if { ! [info exist param_(NamTraceFile)] } {
            puts "CEAS_scenario: Warning - No namtrace file specified. Disabling namtrace."
            set param_(NAMTRACE_ON) 0
        } else {
            set nam_f_ [open $param_(NamTraceFile) w]
            $ns_ namtrace-all $nam_f_
        }
    }
}

# Close/end simulator output
CEAS_scenario instproc close_trace {} {
    $self instvar ns_ param_ trace_f_ nam_f_
    $ns_ flush-trace
    if { [ info exist trace_f_ ] } { close $trace_f_ }
    if { [ info exist nam_f_ ] } { close $nam_f_ }
} 

# Get and setup colors
CEAS_scenario instproc get_colors {} {
    $self instvar ns_ colors_
    global CEASPath
    source $CEASPath/libs/common/src/colors.tcl
    set colors_ $colors
    # Setup ns color classes
    set i 0
    foreach c $colors_ {
        $ns_ color $i $c
        incr i
    }
}

# Show progress info
CEAS_scenario instproc show_progress {} {
    $self instvar ns_ SimID_ param_
    if { [info exist param_(ProgressPeriod)] } {
        if { [$ns_ is-started] } {
            # Simulator running. Output progress info.
            set now [$ns_ now]
            puts [format "$SimID_: Time %.1f" $now]
            flush stdout
            # Reschedule
            $ns_ at [expr $now + $param_(ProgressPeriod)] "$self show_progress"
        }
    } else {
        puts "(CEAS_scenario: Progress info disabled.)"
    }
}

# Stop current ns process and continue another time period (param(Finish)) when
# restarted
CEAS_scenario instproc stop_and_continue {} {  
    $self instvar next_stop_ param_ ns_ SimID_
    set infotext "To continue: kill -SIGCONT [eval pid]\nTo terminate: kill -SIGKILL [eval pid]"
    if {[info exists param_(Email)]} {
        # Inform operator by email
        exec  echo "$infotext" | mail -s "Simulation $SimID_ ([eval pid]) stopped" $param_(Email) 
    }
    puts "Stopped.\n$infotext"
    flush stdout    
    $ns_ flush-trace
    # Stop this process
    exec kill -SIGSTOP [ eval pid ]
    # Continute another time period
    set next_stop_ [expr $next_stop_ + $param_(Finish)]
    $ns_ at $next_stop_ "$self stop_and_continue"
}

# Clean up and end simulation
CEAS_scenario instproc finish {} {
    $self close_trace
    exit 0
}
