/*
 * Copyright (c) 2008 Q2S NTNU, Trondheim, Norway
 *  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Port of the Swarm module developed by O. Wittner for ns-2.30.
 * Author: Laurent Paquereau <laurent.paquereau@q2s.ntnu.no>
 *
 * $Header$
 */
 
 #include "plain-ceas-routing-unit.h"

//! Implement ant behaviour
void
Plain_CEASRoutingUnit::run()
{
    hdr_cmn* ch = hdr_cmn::access(p_);
    hdr_ip*  ih = hdr_ip::access(p_);
    key_ = ant()->species();

    switch (ant()->state()) {
      
      case CEAS_STATE_IDLE:    
        // change to forward ant
        ant()->L() = -NOW; // birthtime
        idle2Forward();
                
      case CEAS_STATE_FORWARDSEARCH:    
        if (!ant()->isAtDestination(ih->daddr(),local_address()))
          { // ant is not at destination
            // get list of not already visited neighbour
            vector<pair<NeighbourEntry*, double> > neighbours;
            networkLayerUnit()->getNeighboursCumulativePheroLevel(
              species(),          // species
              ANY_ADDRESS,        // source
              ih->daddr(),        // destination
              neighbours,         // list of unvisited neighbours
              ant()->behaviour(), // ignore pherolevel 0(normal)/1(explorer)
              ant()->path());     // blacklist
            if (neighbours.empty())
              // all neighbours have been visited. allow to revisit any 
              networkLayerUnit()->getAllNeighbours(neighbours);
            // select which neighbour to visit next by applying phero distr
            NeighbourEntry* next =
              networkLayerUnit()->selectRandomNeighbour(neighbours);
            // record visit to current node and cost to next
            ant()->push(local_address(),next);
            ant()->L() -= next->costTo(); // cancel what is done in ceas-packet
            // move to next
            sendTo(next);
            break;
          }
        else
          { // ant is at destination
            // compute path cost (total delay experienced by the ant)
            ant()->L() += NOW;
            // record visit to current node
            ant()->push(local_address(),0);

            if (routingLoops_ == CEAS_RT_LOOPS_REMOVE_AT_DESTINATION)
              // remove loops from the path
              removeLoops(&ant()->path());
            
            // do rho reduction
            // doRhoReduction();
            // update temperature for all ants
            double gamma = updateTotalTemperature();
            assert(gamma>0);
            // do elite selection
            bool backtrack = doEliteSelection(gamma);
            ant()->gamma() = gamma;
            assert(!backtrack || gamma>0);
            if (logger()) logger()->sample(this,CEAS_LOG_COST_ALL|CEAS_LOG_PATH_ALL);
            if (!backtrack)
              { // discard the packet
                Packet::free(p_);
                return;
              }
            if (logger()) logger()->sample(this,CEAS_LOG_COST|CEAS_LOG_PATH);
            // backtrack
            forward2Backtracking();
          }
        
      case CEAS_STATE_BACKTRACKING:{              
        assert(ant()->path().back().first == local_address());
        if (!ant()->isAtDestination(ih->saddr(),local_address()))
          { // not at destination
            networkLayerUnit()->updatePhero(
              species(),                         // species
              ANY_ADDRESS,                       // source address
              ih->saddr(),                       // ant destination
              ant()->path().back().second.next(),// neighbor key
              ant()->gamma(),                    // temperature
              ant()->L(),                        // path cost 
              ant()->beta());                    // beta
            if (logger()) logger()->samplePhero(this);
          }
        // remove record of the visit at the current node
        ant()->pop();
        assert(ant()->path().size()==ih->ttl());
        if (ant()->isBackAtNest())
          { // backtracking complete, terminate agent
            assert(ih->ttl()==0);
            Packet::free(p_);
            return;
          }
        // move to next
        vector<NeighbourEntry*> neighbours;
        networkLayerUnit()->getNeighbours(
          ant()->path().back().first, // next node address
          ant()->path().back().second.next().getChannelIndex(), // fwd:channel
          neighbours);
        if (neighbours.empty())
          { // neighbour not found or not reachable, drop the packet
            Packet::free(p_);
            return;
          }
        NeighbourEntry* next = neighbours.front();
        sendTo(next);
        break;
      }
      default:
        fprintf(stderr,"Plain_CEASRoutingUnit::%s: Invalid state "  \
                "%d\n",__FUNCTION__,ant()->state());
        abort();
        break;
    }
}
