
#### Automatic detection of the path to CEAS sources ####

set nsPath [exec which ns]
set CEASPath [string replace $nsPath [expr [string length $nsPath] - 6] [expr [string length $nsPath] -1] CEAS]

#### Parameters ####

set param(DEBUG_ON)          0
set param(TRACE_ON)          1
set param(NAMTRACE_ON)       1
set param(Topology)          "uninett.tcl"
#set param(Topology)          "q2s_topo_veb_ten_nodes.tcl"   ;# 10 node test of pheromone sharing 
set param(StopToContinue)    0        ;# Flag indicating if simulation is to stop not terminate
set param(Finish)            300
set param(ProgressPeriod)   [expr $param(Finish) / 1000 ]  ;# Time periode for progress report
if { $param(ProgressPeriod) < 1  } {
    set param(ProgressPeriod) 1
}
set param(BinSize)           100                    ;# No of samples in bins when averaging
set param(SampleFreq)        2
set param(GenRate)           1.0       ;# No of swarm agents per second to generate
set param(ExplorerRate)      0.1               ;# No of explorer swarm agents per second to generate
set param(InitPhase)       [expr 10 / $param(GenRate) ]
# Matrix of connections, a row is "species sources-destination"
set param(Connections)       "
1 0-9
2 1-9
3 2-9
4 3-9"
set param(TTL)               130      ;# Initial ttl value
set param(ProcDelayAnt)      0.000    ;# CEAS agent processing delay in nodes
set param(ProcDelayData)     0.000    ;# Data packet processing delay in nodes
set param(Beta)              0.998     ;# "Evaporation" rate (path memory)
set param(Rho)               0.01      ;# Search focus
set param(TraceFile)         "trace.tr"
set param(NamTraceFile)      "namtrace.nam"

#### Scenario Classes ####

# Base class
source $CEASPath/libs/common/src/ceas-scenario.tcl

# Load dynamic library
dynlibload ceasplain

# CURRENT SCENARIO
Class Plain_scenario -superclass CEAS_scenario

Plain_scenario instproc init {param} {
    # Constructor
    $self instvar SimID_
    # Add SimID to param array
    array set param_tmp $param
    $self get_simid
    set param_tmp(SimID) $SimID_ ;
    puts "SimID:"
    puts $SimID_
    set param [array get param_tmp]
    # Add scenario title to param-array and run parent class constructor
    $self next "$param Scenario Plain_scenario"
}

Plain_scenario instproc init_simulation {} {
    $self next
    $self instvar ns_ param_

    # Configure CEAS
    $ns_ CEAS-config -elite-selection   non-explorer \
                     -routing-loops     remove-at-destination \
                     -log-bin-size      $param_(BinSize) \
                     -log-sample-freq   $param_(SampleFreq) \
                     -log-only 1
    $ns_ dump-CEAS-config
}

Plain_scenario instproc init_components {} {
    # Prepare misc. component required in simulation
    $self instvar gen_ ns_ node_ param_ gen_expl_ ceas_network_layer_unit_

    # create random generator (common for all swarm systems)
    set randomgen [new RNG]
    for {set r 0} {$r < $param_(SimID)} {incr r} {
        $randomgen next-substream
    }

    # Set all queue lengths to 1
    set i [array startsearch node_ ]
    while { [ array anymore node_ $i ] } {
	set ix [array nextelement node_ $i]
	set j [array startsearch node_ ]
	while { [ array anymore node_ $j ] } {
	    set jx [array nextelement node_ $j]
	    set link [ $ns_ link $node_($ix) $node_($jx) ]
	    if { $link != "" } {
		# Link exists between nodes. Set queue limit
		$ns_ queue-limit $node_($ix) $node_($jx) 5000
	     }
	}
	array donesearch node_ $j
    }
    array donesearch node_ $i




    # Get array of connections
    array set connection $param_(Connections)

    # Create and init CEAS packet generators, one per connections
    set i [ array startsearch connection ]
    while { [ array anymore connection $i ] } {

	set species [ array nextelement connection $i]
	set src [ lindex [split $connection($species) "-"] 0]
	set dst [ lindex [split $connection($species) "-"] 1]
	
	puts "id $species, src $src, dst $dst"

	set gen_($species) [$ceas_network_layer_unit_($src) create-generator CEAS/Plain]
	$gen_($species) set-type CBR
	$gen_($species) set-rate $param_(GenRate)
	$gen_($species) set-species $species
	$gen_($species) set-beta $param_(Beta)
	$gen_($species) set ttl_ $param_(TTL)
	$gen_($species) set-dst-address [$node_($dst) node-addr]

	# Explorer generator
	set gen_expl_($species) [$ceas_network_layer_unit_($src) create-generator CEAS/Plain]
	$gen_expl_($species) set-type CBR
	$gen_expl_($species) set-dst-address [$node_($dst) node-addr]
	$gen_expl_($species) set-behaviour explorer
	$gen_expl_($species) set-rate $param_(ExplorerRate)
	$gen_expl_($species) set-species $species
	$gen_expl_($species) set-beta $param_(Beta)
	$gen_expl_($species) set ttl_ $param_(TTL)

    }
    array donesearch connection $i


    # For each node...
    set n_idx [array startsearch node_]
    while {[array anymore node_ $n_idx]} {
	set n [array nextelement node_ $n_idx]

        # Following line is needed when ForwardingUnit installed on a node
	[$node_($n) network-layer-manager] default-target $ceas_network_layer_unit_($n)

	# Set the random generator
        $ceas_network_layer_unit_($n) use-rng $randomgen

	# Set swarm agent processing time 
	$ceas_network_layer_unit_($n) set-service-time $param_(ProcDelayAnt)
	$ceas_network_layer_unit_($n) set-service-time-data $param_(ProcDelayData)

	# Attach logger
	$ceas_network_layer_unit_($n) attach-logger

 	}

  array donesearch node_ $n_idx
	
}

Plain_scenario instproc schedule_events {} {
    # List events to take place during the sim. scenario
    # Schedule events
    $self next                ;# (Call parent class instproc)
    $self instvar ns_ param_ node_ numnodes_ colors_ gen_  gen_expl_

    # Get array of connections
    array set connection $param_(Connections)

    # Schedule start of  swarm generators
    set i [ array startsearch connection ]
    while { [ array anymore connection $i ] } {
	set species [ array nextelement connection $i]

	$ns_ at 0.0 "$gen_($species) set-behaviour explorer"   
	$ns_ at 0.0 "$gen_($species) start"   
	$ns_ at $param_(InitPhase) "$gen_($species) set-behaviour normal"   
	$ns_ at 0.0 "$gen_expl_($species) start" 
	
    }

#$ns_ rtmodel-at 10000.0 down $node_(6) $node_(9)
#$ns_ rtmodel-at 17000.0 up $node_(6) $node_(9)
	
	array donesearch connection $i
}     
    
###########
# M A I N #  
###########

# Include only swarm packets and IP packets
remove-all-packet-headers 
add-packet-header IP CEAS

#
# Create simulator instance
#
set ns [new Simulator] 

# Create scenario
new Plain_scenario [array get param]

# Run simulation
$ns run
